
/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Consistency check on the local geometry arrays bb.ipp,bb.imb and of the
* gauge field arrays bb.u,bb.ud of the boundary of a new block b
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "start.h"
#include "misc.h"
#include "block.h"
#include "global.h"

static int ix_test[VOLUME];


static int check_unity(su3 *u)
{
   int i;
   float *rv;
   su3 v;

   v=*u;

   v.c11.re-=1.0f;
   v.c22.re-=1.0f;
   v.c33.re-=1.0f;

   rv=(float*)(&v);

   for (i=0;i<18;i++)
      if (rv[i]!=0.0f)
         return 1;

   return 0;
}


static int check_unity_dble(su3_dble *u)
{
   int i;
   double *rv;
   su3_dble v;

   v=*u;

   v.c11.re-=1.0;
   v.c22.re-=1.0;
   v.c33.re-=1.0;

   rv=(double*)(&v);

   for (i=0;i<18;i++)
      if (rv[i]!=0.0)
         return 1;

   return 0;
}


int main(int argc,char *argv[])
{
   int my_rank;
   int bo[4],bs[4];
   int ix,iy,iz,is,mu,ifc,itest;
   su3 *ub;
   su3_dble *udb;
   block_t b;
   bndry_t *bb;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check2.log","w",stdout);
      fin=freopen("check2.in","r",stdin);

      printf("\n");
      printf("Consistency checks on the local arrays of the block boundary\n");
      printf("------------------------------------------------------");
      printf("------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bo","%d %d %d %d\n",&bo[0],&bo[1],&bo[2],&bo[3]);
      read_line("bs","%d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);
      fclose(fin);

      printf("bo = %d %d %d %d\n",bo[0],bo[1],bo[2],bo[3]);
      printf("bs = %d %d %d %d\n",bs[0],bs[1],bs[2],bs[3]);
   }

   MPI_Bcast(bo,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);

   start_ranlux(0,1234);
   geometry();
   alloc_u();
   alloc_ud();

   alloc_blk(&b,bo,bs,1,1,0,0);
   alloc_bnd(&b,0,0);
   bb=b.bb;

   for (ifc=0;ifc<8;ifc++)
   {
      mu=ifc/2;

      error(((*bb).ifc!=ifc)||((*bb).vol!=(b.vol/bs[mu])),1,
            "main [check2.c]","bb.ifc or bb.vol is incorrect");

      error(((*bb).ns!=0)||((*bb).nsd!=0)||((*bb).s!=NULL)||((*bb).sd!=NULL),1,
            "main [check2.c]",
            "The spinor field parameters are incorrectly set");

      for (ix=0;ix<VOLUME;ix++)
         ix_test[ix]=0;

      itest=0;

      for (ix=0;ix<(*bb).vol;ix++)
      {
         iy=(*bb).ipp[ix];

         if ((iy<0)||(iy>=b.vol))
            itest=1;

         iy=(*bb).imb[ix];

         if ((iy<0)||(iy>=VOLUME))
            itest=2;
         else
         {
            ix_test[iy]+=1;

            if (ix_test[iy]>1)
               itest=3;
         }
      }

      error(itest==1,1,"main [check2.c]",
            "ipp is out of range");
      error(itest==2,1,"main [check2.c]",
            "imb is out of range");
      error(itest==3,1,"main [check2.c]",
            "imb is not one-to-one");

      for (ix=0;ix<((*bb).vol/2);ix++)
      {
         iy=(*bb).ipp[ix];
         iz=(*bb).ipp[ix+((*bb).vol/2)];

         if ((iy<b.volh)||(iz>=b.volh))
            itest=1;

         iy=(*bb).imb[ix];
         iz=(*bb).imb[ix+(*bb).vol/2];

         if ((iy>=(VOLUME/2))||(iz<(VOLUME/2)))
            itest=2;
      }

      error(itest==1,1,"main [check2.c]",
            "ipp does not respect the even-odd ordering");
      error(itest==2,1,"main [check2.c]",
            "imb does not respect the even-odd ordering");

      for (ix=0;ix<(*bb).vol;ix++)
      {
         iy=(*bb).ipp[ix];
         is=0;

         if (((ifc%2)==0)&&(b.idn[iy][mu]==b.vol))
            is=1;

         if (((ifc%2)==1)&&(b.iup[iy][mu]==b.vol))
            is=1;

         if (is==0)
            itest=1;
      }

      error(itest==1,1,"main [check2.c]",
            "Partner points are not on the interior boundary of the block");

      for (ix=0;ix<(*bb).vol;ix++)
      {
         iy=(*bb).ipp[ix];
         iz=b.imb[iy];
         iy=(*bb).imb[ix];

         if ((ifc%2)==0)
            iy=iup[iy][mu];
         else
            iy=idn[iy][mu];

         if ((*bb).ibn==1)
            iy=map[iy-VOLUME];

         if (iy!=iz)
            itest=1;
      }

      error(itest==1,1,"main [check2.c]",
            "Embedding of the boundary is incorrect");

      for (ix=0;ix<(*bb).vol;ix++)
      {
         ub=(*bb).u+ix;
         udb=(*bb).ud+ix;

         if (check_unity(ub)!=0)
            itest=1;

         if (check_unity_dble(udb)!=0)
            itest=2;
      }

      error(itest==1,1,"main [check2.c]",
            "The gauge field b.u is not properly initialized");
      error(itest==2,1,"main [check2.c]",
            "The gauge field b.ud is not properly initialized");
      bb+=1;
   }

   error_chk();

   if (my_rank==0)
   {
      printf("\n");
      printf("The block boundary is correctly mapped by the local arrays\n\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
