
/*******************************************************************************
*
* File time3.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* BiCGstab, SAP+GCR and deflated SAP+GCR timing
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <float.h>
#include "mpi.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "flags.h"
#include "block.h"
#include "linalg.h"
#include "linsolv.h"
#include "sw_term.h"
#include "dirac.h"
#include "eva.h"
#include "sap_gcr.h"
#include "dfl.h"
#include "global.h"

typedef struct
{
   double kappa;
   int nmx_bcg,nmx_sap,nmx_dfl;
   int nkv_sap,nkv_dfl;
} kparms_t;

typedef struct
{
   int nfail;
   int nmin,nmax,navg;
   double tmin,tmax,tavg;
   double rmax;
} data_t;

static int my_rank,id;
static int bc,nc_first,nc_last;
static int bs_sap[4],nmr,ncy;
static int bs_dfl[4],Ns,rinv,rnmr,rncy,rvnmx;
static int vnkv,vnmx,vdnmx,nk;
static double beta,kappa,csw,res;
static double rkappa,rvres,vres,vdres;
static char cnfg_dir[NAME_SIZE];
static kparms_t *kparms;
static data_t *bcgdat,*sapdat,*dfldat,*mdsdat,*zopdat;
static data_t *litdat,*lisdat,*liddat;


static void alloc_data(void)
{
   int ik;
   data_t *dat;
   
   kparms=amalloc(nk*sizeof(*kparms),3);
   bcgdat=amalloc((6*nk+2)*sizeof(*bcgdat),3);

   error((kparms==NULL)||(bcgdat==NULL),1,"alloc_data [time3.c]",
         "Unable to allocate data arrays");

   sapdat=bcgdat+nk;
   dfldat=sapdat+nk;
   litdat=dfldat+nk;
   lisdat=litdat+nk;
   liddat=lisdat+nk;
   mdsdat=liddat+nk;
   zopdat=mdsdat+1;
   dat=bcgdat;

   for (ik=0;ik<(6*nk+2);ik++)
   {
      (*dat).nfail=0;
      (*dat).nmin=INT_MAX;
      (*dat).nmax=0;
      (*dat).navg=0;
      (*dat).tmin=DBL_MAX;      
      (*dat).tmax=0.0;
      (*dat).tavg=0.0;
      (*dat).rmax=0.0;

      dat+=1;
   }
}


static void read_kparms(int ik)
{
   int nmx_bcg,nmx_sap,nmx_dfl;
   int nkv_sap,nkv_dfl;
   double kappa_val;
   
   if (my_rank==0)
   {
      read_line("kappa_val","%lf\n",&kappa_val);      
      read_line("nmx_bicg","%d\n",&nmx_bcg);
      read_line("nmx_sap","%d\n",&nmx_sap);
      read_line("nmx_dfl_sap","%d\n",&nmx_dfl);
      read_line("nkv_sap","%d\n",&nkv_sap);
      read_line("nkv_dfl_sap","%d\n",&nkv_dfl);      
   }

   MPI_Bcast(&kappa_val,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_bcg,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_dfl,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nkv_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nkv_dfl,1,MPI_INT,0,MPI_COMM_WORLD);

   kparms[ik].kappa=kappa_val;
   kparms[ik].nmx_bcg=nmx_bcg;
   kparms[ik].nmx_sap=nmx_sap;
   kparms[ik].nmx_dfl=nmx_dfl;   
   kparms[ik].nkv_sap=nkv_sap;
   kparms[ik].nkv_dfl=nkv_dfl;   
}


static void read_infile(void)
{
   int ik;
   FILE *fin=NULL;
      
   if (my_rank==0)
   {
      fin=freopen("time3.in","r",stdin);

      read_line("id","%d\n",&id);
      read_line("cnfg_dir","%s\n",cnfg_dir);
      read_line("beta","%lf\n",&beta);
      read_line("kappa","%lf\n",&kappa);
      read_line("csw","%lf\n",&csw);      
      read_line("bc","%d\n",&bc);      
      read_line("nc_first","%d\n",&nc_first);
      read_line("nc_last","%d\n",&nc_last);  
      read_line("res","%lf\n",&res);

      read_line("bs_sap","%d %d %d %d\n",
                &bs_sap[0],&bs_sap[1],&bs_sap[2],&bs_sap[3]);
      read_line("nmr","%d\n",&nmr);      
      read_line("ncy","%d\n",&ncy);      

      read_line("bs_dfl","%d %d %d %d\n",
                &bs_dfl[0],&bs_dfl[1],&bs_dfl[2],&bs_dfl[3]);
      read_line("Ns","%d\n",&Ns);
      read_line("rkappa","%lf\n",&rkappa);
      read_line("rinv","%d\n",&rinv);
      read_line("rnmr","%d\n",&rnmr);
      read_line("rncy","%d\n",&rncy);
      read_line("rvnmx","%d\n",&rvnmx);      
      read_line("rvres","%lf\n",&rvres);
      read_line("vnkv","%d\n",&vnkv);
      read_line("vnmx","%d\n",&vnmx);
      read_line("vres","%lf\n",&vres);
      read_line("vdnmx","%d\n",&vdnmx);
      read_line("vdres","%lf",&vdres);
      read_line("nk","%d\n",&nk);      
   }

   MPI_Bcast(&id,1,MPI_INT,0,MPI_COMM_WORLD);      
   MPI_Bcast(cnfg_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(&beta,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&kappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&csw,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_first,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_last,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&res,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   
   MPI_Bcast(bs_sap,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmr,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncy,1,MPI_INT,0,MPI_COMM_WORLD);

   MPI_Bcast(bs_dfl,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&Ns,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&rkappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&rinv,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&rnmr,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&rncy,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&rvnmx,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&rvres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&vnkv,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&vnmx,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&vres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&vdnmx,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&vdres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nk,1,MPI_INT,0,MPI_COMM_WORLD);

   alloc_data();

   for (ik=0;ik<nk;ik++)
      read_kparms(ik);

   if (my_rank==0)
      fclose(fin);      
}


static void print_parms(void)
{
   int ik;
   
   if (my_rank==0)
   {
      printf("id = %d\n",id);
      printf("beta = %.6f\n",beta);
      printf("kappa = %.6f\n",kappa);
      printf("csw = %.6f\n",csw);
      printf("bc = %+d\n",bc);
      printf("res = %.1e\n",res);       
      
      printf("bs_sap = %d %d %d %d\n",
             bs_sap[0],bs_sap[1],bs_sap[2],bs_sap[3]);
      printf("nmr = %d\n",nmr);
      printf("ncy = %d\n",ncy);
      
      printf("bs_dfl = %d %d %d %d\n",
             bs_dfl[0],bs_dfl[1],bs_dfl[2],bs_dfl[3]);
      printf("Ns = %d\n",Ns);
      printf("rkappa = %.6f\n",rkappa);       
      printf("rinv = %d\n",rinv);
      printf("rnmr = %d\n",rnmr);       
      printf("rncy = %d\n",rncy);
      printf("rvnmx = %d\n",rvnmx);
      printf("rvres = %.1e\n",rvres);       
      printf("vnkv = %d\n",vnkv);
      printf("vnmx = %d\n",vnmx);
      printf("vres = %.1e\n",vres);
      printf("vdnmx = %d\n",vdnmx);
      printf("vdres = %.1e\n\n",vdres);

      for (ik=0;ik<nk;ik++)
      {
         printf("kappa_val = %.6f\n",kparms[ik].kappa);
         printf("nmx_bicg = %d\n",kparms[ik].nmx_bcg);
         printf("nmx_sap = %d\n",kparms[ik].nmx_sap);
         printf("nmx_dfl_sap = %d\n",kparms[ik].nmx_dfl);         
         printf("nkv_sap = %d\n",kparms[ik].nkv_sap);
         printf("nkv_dfl_sap = %d\n\n",kparms[ik].nkv_dfl);
      }
      
      printf("Configurations no %d -> %d\n\n",nc_first,nc_last);
      fflush(stdout);      
   }
}


static void mdstim(void)
{
   int status,ifail;
   double wt1,wt2,dwt;

   set_gcr_parms(bs_sap,1,1,1);
   sw_term();
   assign_ud2u();
   assign_swd2sw();
   assign_ud2ubgr(GCR_BLOCKS);
   
   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();       
   dfl_modes(0,&status,&ifail);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();

   dwt=wt2-wt1;

   if ((status<0)||(ifail!=0))
      mdsdat[0].nfail+=1;
   else
   {
      if (mdsdat[0].nmin>status)
         mdsdat[0].nmin=status;
      if (mdsdat[0].nmax<status)
         mdsdat[0].nmax=status;      

      mdsdat[0].navg+=status;            

      if (mdsdat[0].tmin>dwt)
         mdsdat[0].tmin=dwt;
      if (mdsdat[0].tmax<dwt)
         mdsdat[0].tmax=dwt;      

      mdsdat[0].tavg+=dwt;
   }
   
   if (my_rank==0)
      printf("dfl_modes:    status = %4d, ifail = %d, time = %.2e sec\n",
             status,ifail,dwt);
}


static void zoptim(void)
{
   double wt1,wt2,dwt;

   set_flags(NEW_DFLMODES);
   sw_term();

   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();
   zopnohat();
   zophat();
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();
   dwt=wt2-wt1;

   if (zopdat[0].tmin>dwt)
      zopdat[0].tmin=dwt;
   if (zopdat[0].tmax<dwt)
      zopdat[0].tmax=dwt;      

   zopdat[0].tavg+=dwt;
   
   if (my_rank==0)
      printf("Z matrices:                             time = %.2e sec\n\n",dwt);
}


static double residue(spinor_dble *s,spinor_dble *r)
{
   int ir,iw;
   double d;
   complex_dble z;
   spinor_dble **wsd;
   
   if ((query_flags(SWD_UP2DATE)!=1)||
       (query_flags(SWD_E_INVERTED)!=0)||(query_flags(SWD_O_INVERTED)!=0))
   {
      sw_term();
   }
   
   wsd=reserve_wsd(1);

   ir=(r-psd[0][0])/NSPIN;   
   iw=(wsd[0]-psd[0][0])/NSPIN;

   Qnohat_dble(ir,iw);
   mulg5_dble(VOLUME,wsd[0]);
   z.re=-1.0;
   z.im=0.0;
   mulc_spinor_add_dble(VOLUME,wsd[0],s,z);
   d=norm_square_dble(VOLUME,1,wsd[0]);

   release_wsd();

   return sqrt(d);
}


static void Ophat_dble(spinor_dble *pk,spinor_dble *pl)
{
   int k,l;

   k=(pk-psd[0][0])/NSPIN;
   l=(pl-psd[0][0])/NSPIN;

   Qhat_dble(k,l);
   mulg5_dble(VOLUME/2,pl);
}


static void bcgtim(int ik,spinor_dble *s)
{
   int status,ifail,iw5,iw6;
   double wt1,wt2,dwt,nres;
   complex_dble z;
   spinor_dble **wsd;

   sw_term();
   ifail=invert_swd(ODD_PTS);

   if (ifail==0)
   {
      wsd=reserve_wsd(7);
      iw5=(wsd[5]-psd[0][0])/NSPIN;
      iw6=(wsd[6]-psd[0][0])/NSPIN;
      
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();

      assign_sd2sd(VOLUME,s,wsd[5]);
      mulg5_dble(VOLUME,wsd[5]);
      apply_sw_dble(VOLUME/2,swd+VOLUME,wsd[5]+(VOLUME/2),wsd[5]+(VOLUME/2));
      Qeo_dble(iw5,iw5);
      mulg5_dble(VOLUME/2,wsd[5]);
      
      bicg(VOLUME/2,1,Ophat_dble,wsd,kparms[ik].nmx_bcg,res,
           wsd[5],wsd[6],&status);

      Qoe_dble(iw6,iw6);
      apply_sw_dble(VOLUME/2,swd+VOLUME,wsd[6]+(VOLUME/2),wsd[6]+(VOLUME/2));
      z.re=-1.0;
      z.im=0.0;
      mulc_spinor_add_dble(VOLUME/2,wsd[5]+(VOLUME/2),wsd[6]+(VOLUME/2),z);
      assign_sd2sd(VOLUME/2,wsd[5]+(VOLUME/2),wsd[6]+(VOLUME/2));
      
      MPI_Barrier(MPI_COMM_WORLD);
      wt2=MPI_Wtime();
      dwt=wt2-wt1;

      if (status<0)
      {
         bcgdat[ik].nfail+=1;
         nres=1.0;
      }
      else
      {
         if (bcgdat[ik].nmin>status)
            bcgdat[ik].nmin=status;
         if (bcgdat[ik].nmax<status)
            bcgdat[ik].nmax=status;      

         bcgdat[ik].navg+=status;            

         if (bcgdat[ik].tmin>dwt)
            bcgdat[ik].tmin=dwt;
         if (bcgdat[ik].tmax<dwt)
            bcgdat[ik].tmax=dwt;      

         bcgdat[ik].tavg+=dwt;

         nres=residue(s,wsd[6]);

         if (nres>bcgdat[ik].rmax)
            bcgdat[ik].rmax=nres;
      }

      release_wsd();
   }
   else
   {
      bcgdat[ik].nfail+=1;
      status=0;
      dwt=0.0;
      nres=1.0;
   }
   
   if (my_rank==0)
      printf("BiCGstab:     status = %4d, ifail = %d, time = %.2e sec, "
             "residue = %.1e\n",status,ifail,dwt,nres);
}


static void saptim(int ik,spinor_dble *s)
{
   int status,ifail,is,iw;
   double wt1,wt2,dwt,nres;
   spinor_dble **wsd;

   set_gcr_parms(bs_sap,nmr,ncy,kparms[ik].nkv_sap);
   sw_term();
   assign_ud2ubgr(GCR_BLOCKS);
   ifail=assign_swd2swbgr(GCR_BLOCKS,ODD_PTS);

   if (ifail==0)
   {
      wsd=reserve_wsd(1);
      is=(s-psd[0][0])/NSPIN;
      iw=(wsd[0]-psd[0][0])/NSPIN;
      
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();
      sap_gcr(kparms[ik].nmx_sap,res,is,iw,&status,&ifail);
      MPI_Barrier(MPI_COMM_WORLD);
      wt2=MPI_Wtime();
      dwt=wt2-wt1;

      if (status<0)
      {
         sapdat[ik].nfail+=1;
         nres=1.0;
      }
      else
      {
         if (sapdat[ik].nmin>status)
            sapdat[ik].nmin=status;
         if (sapdat[ik].nmax<status)
            sapdat[ik].nmax=status;      

         sapdat[ik].navg+=status;            

         if (sapdat[ik].tmin>dwt)
            sapdat[ik].tmin=dwt;
         if (sapdat[ik].tmax<dwt)
            sapdat[ik].tmax=dwt;      

         sapdat[ik].tavg+=dwt;

         nres=residue(s,wsd[0]);

         if (nres>sapdat[ik].rmax)
            sapdat[ik].rmax=nres;
      }

      release_wsd();
   }
   else
   {
      sapdat[ik].nfail+=1;
      status=0;
      dwt=0.0;
      nres=1.0;
   }
   
   if (my_rank==0)
      printf("sap_gcr:      status = %4d, ifail = %d, time = %.2e sec, "
             "residue = %.1e\n",status,ifail,dwt,nres);
}


static void dfltim(int ik,spinor_dble *s)
{
   int stat[3],ifail,nb,isw,is,iw;
   double wt1,wt2,dwt,nres;
   complex **wv;
   spinor_dble **wsd;

   set_gcr_parms(bs_sap,nmr,ncy,kparms[ik].nkv_dfl);
   sw_term();
   assign_ud2u();
   assign_swd2sw();
   assign_ud2ubgr(GCR_BLOCKS);
   ifail=assign_swd2swbgr(GCR_BLOCKS,ODD_PTS);

   if (ifail==0)
   {
      wv=reserve_wv(2);
      blk_list(DFL_BLOCKS,&nb,&isw);
      random_v(nb*Ns,wv[0],1.0f);
      Znohat(wv[0],wv[1]);
      Zhat(wv[0],wv[1]);
      release_wv();

      wsd=reserve_wsd(1);
      is=(s-psd[0][0])/NSPIN;
      iw=(wsd[0]-psd[0][0])/NSPIN;
      
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();
      dfl_sap_gcr(kparms[ik].nmx_dfl,res,is,iw,stat,&ifail);
      MPI_Barrier(MPI_COMM_WORLD);
      wt2=MPI_Wtime();
      dwt=wt2-wt1;

      if (stat[0]<0)
      {
         dfldat[ik].nfail+=1;
         nres=1.0;
      }
      else
      {
         if (dfldat[ik].nmin>stat[0])
            dfldat[ik].nmin=stat[0];
         if (dfldat[ik].nmax<stat[0])
            dfldat[ik].nmax=stat[0];      

         dfldat[ik].navg+=stat[0];            

         if (dfldat[ik].tmin>dwt)
            dfldat[ik].tmin=dwt;
         if (dfldat[ik].tmax<dwt)
            dfldat[ik].tmax=dwt;      

         dfldat[ik].tavg+=dwt;

         nres=residue(s,wsd[0]);

         if (nres>dfldat[ik].rmax)
            dfldat[ik].rmax=nres;

         if (lisdat[ik].nmin>stat[1])
            lisdat[ik].nmin=stat[1];
         if (lisdat[ik].nmax<stat[1])
            lisdat[ik].nmax=stat[1];      

         lisdat[ik].navg+=stat[1];            

         if (liddat[ik].nmin>stat[2])
            liddat[ik].nmin=stat[2];
         if (liddat[ik].nmax<stat[2])
            liddat[ik].nmax=stat[2];      

         liddat[ik].navg+=stat[2];            
      }

      release_wsd();
   }
   else
   {
      dfldat[ik].nfail+=1;
      stat[0]=0;
      stat[1]=0;
      stat[2]=0;
      dwt=0.0;
      nres=1.0;
   }
   
   if (my_rank==0)
      printf("dfl_sap_gcr:  status = %4d, ifail = %d, time = %.2e sec, "
             "residue = %.1e (little: %3d, %3d)\n",
             stat[0],ifail,dwt,nres,stat[1],stat[2]);
}


static void littim(int ik)
{
   int nb,Ns,nv,status,n,count;
   double wt1,wt2,dwt,nres;
   complex_dble **wvd,z;
   zop_dble_t *zop;

   zop=zopnohat_dble();
   nb=(*zop).nb;
   Ns=(*zop).Ns;   
   nv=nb*Ns;
   
   wvd=reserve_wvd(3);
   random_vd(nv,wvd[0],1.0);
   assign_vd2vd(nv,wvd[0],wvd[2]);
   dfl_solve_little(vdnmx,vdres,wvd[0],&status);

   if (status<0)
   {
      litdat[ik].nfail+=1;
      nres=1.0;
      dwt=0.0;
   }
   else
   {
      if (litdat[ik].nmin>status)
         litdat[ik].nmin=status;
      if (litdat[ik].nmax<status)
         litdat[ik].nmax=status;      

      litdat[ik].navg+=status;  

      Znohat_dble(wvd[0],wvd[1]);
      z.re=-1.0;
      z.im=0.0;
      mulc_vadd_dble(nv,wvd[1],wvd[2],z);
      nres=vnorm_square_dble(nv,1,wvd[1])/vnorm_square_dble(nv,1,wvd[2]);
      nres=sqrt(nres);

      if (nres>litdat[ik].rmax)
         litdat[ik].rmax=nres;
      
      n=(int)(1.0e9/(64.0*(double)(Ns*nv*status)));
      if (n<1)
         n=1;
      
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();
      for (count=0;count<n;count++)
      {
         assign_vd2vd(nv,wvd[2],wvd[0]);
         dfl_solve_little(vdnmx,vdres,wvd[0],&status);
      }
      MPI_Barrier(MPI_COMM_WORLD);
      wt2=MPI_Wtime();
      dwt=(wt2-wt1)/(double)(n);

      if (litdat[ik].tmin>dwt)
         litdat[ik].tmin=dwt;
      if (litdat[ik].tmax<dwt)
         litdat[ik].tmax=dwt;      

      litdat[ik].tavg+=dwt;
   }

   release_wvd();
   
   if (my_rank==0)
      printf("solve_little: status = %4d,            time = %.2e sec, "
             "residue = %.1e\n",status,dwt,nres);
}


int main(int argc,char *argv[])
{
   int nsize,icnfg,idfl,ik,nkv,n;
   spinor_dble **wsd;
   char cnfg_file[NAME_SIZE],nbase[NAME_SIZE];   
   FILE *flog=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("time3.log","w",stdout);
      
      printf("\n");
      printf("BiCGstab, SAP+GCR and deflated SAP+GCR timing\n");
      printf("---------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);
   }

   read_infile();
   print_parms();
   
   start_ranlux(0,123456);   
   geometry();
   alloc_u();
   alloc_ud();
   alloc_sw();
   alloc_swd();

   idfl=0;
   nkv=0;

   for (ik=0;ik<nk;ik++)
   {
      if ((kparms[ik].nmx_sap>0)&&(kparms[ik].nkv_sap>nkv))
         nkv=kparms[ik].nkv_sap;
      if ((kparms[ik].nmx_dfl>0)&&(kparms[ik].nkv_dfl>nkv))
         nkv=kparms[ik].nkv_dfl;
      if (kparms[ik].nmx_dfl>0)
         idfl=1;
   }
   
   if (Ns<(2*nkv))
      alloc_s(2*nkv+1);
   else
      alloc_s(Ns+2);
   alloc_sd(9);

   set_dfl_parms(bs_dfl,Ns,rinv,rnmr,rncy,rvnmx,vnkv,vnmx,vdnmx,
                 rkappa,rvres,vres,vdres,0,3);
   
   check_dir_root(cnfg_dir);
   error_root(nc_first>nc_last,1,"main [time3.c]",
              "Selected range of configurations is void");

   nsize=name_size("%dx%dx%dx%db%.2fk%.5fc%.5fid%d",
                   NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);
   error_root(nsize>=NAME_SIZE,1,"main [time3.c]","Base name is too long");
   sprintf(nbase,"%dx%dx%dx%db%.2fk%.5fc%.5fid%d",
           NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);

   nsize=name_size("%s/%sn%d",cnfg_dir,nbase,nc_last);
   error_root(nsize>=NAME_SIZE,1,"main [time3.c]",
              "cnfg_dir name is too long");

   for (icnfg=nc_first;icnfg<=nc_last;icnfg++)
   {
      message("Configuration no %d:\n\n",icnfg);

      sprintf(cnfg_file,"%s/%sn%d",cnfg_dir,nbase,icnfg);
      import_cnfg(cnfg_file);
      if (bc==-1)
         flipbcd();

      set_lat_parms(beta,kappa,csw);

      if (idfl!=0)
      {
         mdstim();
         zoptim();
      }

      wsd=reserve_wsd(1);
      random_sd(VOLUME,wsd[0],1.0);
      normalize_dble(VOLUME,1,wsd[0]);
      
      for (ik=0;ik<nk;ik++)
      {
         message("kappa_val = %.6f\n",kparms[ik].kappa);
         set_lat_parms(beta,kparms[ik].kappa,csw);

         if (kparms[ik].nmx_bcg>0)
            bcgtim(ik,wsd[0]);
         if (kparms[ik].nmx_sap>0)
            saptim(ik,wsd[0]);
         if (kparms[ik].nmx_dfl>0)
         {
            dfltim(ik,wsd[0]);
            littim(ik);
         }

         if (my_rank==0)
         {
            printf("\n");
            fflush(flog);
         }
      }

      release_wsd();
      error_chk();
   }
   
   if (my_rank==0)
   {
      printf("Timings summary\n");
      printf("---------------\n\n");

      printf("Configurations range %d -> %d\n\n",nc_first,nc_last);      
      
      printf("The three figures quoted in the case of the iteration "
             "numbers (n)\n");
      printf("and the timings (t) are the minimal, maximal and "
             "average values.\n");
      printf("The maximal residue r should be less than "
             "%.1e in all cases.\n\n",res);

      if (idfl!=0)
      {
         n=nc_last-nc_first+1-mdsdat[0].nfail;

         if (n>0)
         {
            printf("Dfl Modes:  nfail = %2d, n = %4d, %4d, %4d, "
                   "t = %.2e, %.2e, %.2e\n",
                   mdsdat[0].nfail,
                   mdsdat[0].nmin,mdsdat[0].nmax,mdsdat[0].navg/n,
                   mdsdat[0].tmin,mdsdat[0].tmax,mdsdat[0].tavg/(double)(n));
         }
         else
            printf("Dfl Modes: program failed in all cases\n");
         
         n=nc_last-nc_first+1;
            
         printf("Z operator:                                   "
                "t = %.2e, %.2e, %.2e\n\n",
                zopdat[0].tmin,zopdat[0].tmax,zopdat[0].tavg/(double)(n));
      }
      
      for (ik=0;ik<nk;ik++)
      {
         printf("kappa_val = %.6f:\n",kparms[ik].kappa);
         
         if (kparms[ik].nmx_bcg>0)
         {
            n=nc_last-nc_first+1-bcgdat[ik].nfail;

            if (n>0)
            {
               printf("BiCGstab:   nfail = %2d, n = %4d, %4d, %4d, "
                      "t = %.2e, %.2e, %.2e, r = %.1e\n",
                      bcgdat[ik].nfail,
                      bcgdat[ik].nmin,bcgdat[ik].nmax,bcgdat[ik].navg/n,
                      bcgdat[ik].tmin,bcgdat[ik].tmax,
                      bcgdat[ik].tavg/(double)(n),bcgdat[ik].rmax);
            }
            else
               printf("BiCGstab: solver failed in all cases\n");
         }

         if (kparms[ik].nmx_sap>0)
         {
            n=nc_last-nc_first+1-sapdat[ik].nfail;

            if (n>0)
            {
               printf("SAP+GCR:    nfail = %2d, n = %4d, %4d, %4d, "
                      "t = %.2e, %.2e, %.2e, r = %.1e\n",
                      sapdat[ik].nfail,
                      sapdat[ik].nmin,sapdat[ik].nmax,sapdat[ik].navg/n,
                      sapdat[ik].tmin,sapdat[ik].tmax,
                      sapdat[ik].tavg/(double)(n),sapdat[ik].rmax);
            }
            else
               printf("SAP+GCR: solver failed in all cases\n");
         }

         if (kparms[ik].nmx_dfl>0)
         {
            n=nc_last-nc_first+1-dfldat[ik].nfail;

            if (n>0)
            {
               printf("Dfl SAP:    nfail = %2d, n = %4d, %4d, %4d, "
                      "t = %.2e, %.2e, %.2e, r = %.1e\n",
                      dfldat[ik].nfail,
                      dfldat[ik].nmin,dfldat[ik].nmax,dfldat[ik].navg/n,
                      dfldat[ik].tmin,dfldat[ik].tmax,
                      dfldat[ik].tavg/(double)(n),dfldat[ik].rmax);

               printf("Little_s:               n = %4d, %4d, %4d\n",
                      lisdat[ik].nmin,lisdat[ik].nmax,lisdat[ik].navg/n);

               printf("Little_sd:              n = %4d, %4d, %4d\n",
                      liddat[ik].nmin,liddat[ik].nmax,liddat[ik].navg/n);
            }
            else
               printf("SAP+GCR: solver failed in all cases\n");

            n=nc_last-nc_first+1-litdat[ik].nfail;

            if (n>0)
            {
               printf("Little_vd:  nfail = %2d, n = %4d, %4d, %4d, "
                      "t = %.2e, %.2e, %.2e, r = %.1e\n",
                      litdat[ik].nfail,
                      litdat[ik].nmin,litdat[ik].nmax,litdat[ik].navg/n,
                      litdat[ik].tmin,litdat[ik].tmax,
                      litdat[ik].tavg/(double)(n),litdat[ik].rmax);
            }
            else
               printf("Little: solver failed in all cases\n");
         }

         printf("\n");
      }

      fclose(flog);
   }
   
   MPI_Finalize();   
   exit(0);
}
