
/*******************************************************************************
*
* File time2b.c
*
* Copyright (C) 2005, 2008 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of Qhat_blk_dble
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "flags.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "sw_term.h"
#include "dirac.h"
#include "block.h"
#include "global.h"


static double wt_Qhat(void)
{
   int my_rank,nmax,n,ib;
   int nb,isw;
   double wt1,wt2,wdt,wtav;
   block_t *b0,*bm,*b;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);   

   b0=blk_list(HMC_BLOCKS,&nb,&isw);
   bm=b0+nb;

   random_sd((*b0).vol,(*b0).sd[0],1.0f);
   random_sd((*b0).vol,(*b0).sd[1],1.0f);
   random_sd((*b0).vol,(*b0).sd[2],1.0f);
   nmax=1;
   
   for (ib=0;ib<1;nmax*=2)
   { 
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();

      for (n=0;n<nmax;n++)
      {
         for (b=b0;b<bm;b++)
         {
            Qhat_blk_dble(b,0,1);
            Qhat_blk_dble(b,1,2);
         }
      }
      
      wt2=MPI_Wtime();
      wdt=wt2-wt1;

      MPI_Reduce(&wdt,&wtav,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);      

      if (my_rank==0)
      {
         wtav/=(double)(NPROC);

         if (wtav>2.0)
            ib=1;

         wtav/=(double)(2*nmax*nb);
      }

      MPI_Bcast(&ib,1,MPI_INT,0,MPI_COMM_WORLD);
   }

   MPI_Bcast(&wtav,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   
   return wtav;
}


int main(int argc,char *argv[])
{
   int my_rank,bs[4],vol;
   double wdt;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("time2b.log","w",stdout);
      fin=freopen("time2b.in","r",stdin);

      printf("\n");
      printf("Timing of Qhat_blk_dble\n");
      printf("-----------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bs","%d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);
      fclose(fin);

      printf("bs = %d %d %d %d\n\n",bs[0],bs[1],bs[2],bs[3]);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);
   vol=bs[0]*bs[1]*bs[2]*bs[3];

   if (my_rank==0)
   {
      if (NPROC>1)
         printf("There are %d MPI processes\n",NPROC);
      else
         printf("There is 1 MPI process\n");

      if ((vol*sizeof(double))<(64*1024))
      {      
         printf("The size of the block gauge field is %d KB\n",
                (int)((72*vol*sizeof(double))/(1024)));
         printf("The size of a block quark field is %d KB\n",
                (int)((24*vol*sizeof(double))/(1024)));
      }
      else
      {
         printf("The size of the block gauge field is %d MB\n",
                (int)((72*vol*sizeof(double))/(1024*1024)));
         printf("The size of a block quark field is %d MB\n",
                (int)((24*vol*sizeof(double))/(1024*1024)));
      }

#if (defined SSE3)
      printf("Using inline assembly SSE3 instructions\n");
#elif (defined SSE2)
      printf("Using inline assembly SSE2 instructions\n");      
#elif (defined SSE)
      printf("Using inline assembly SSE instructions\n");
#endif

#if (defined SSE)
#if (defined P3)
      printf("Assuming SSE prefetch instructions fetch 32 bytes\n");
#elif (defined PM)
      printf("Assuming SSE prefetch instructions fetch 64 bytes\n");
#elif (defined P4)
      printf("Assuming SSE prefetch instructions fetch 128 bytes\n");
#else
      printf("SSE prefetch instructions are not used\n");
#endif
#endif
      
      printf("\n");
   }
   
   start_ranlux(0,1234);
   geometry();
   alloc_ud();
   alloc_swd();

   random_ud();
   set_lat_parms(0.0,0.1,0.2);
   sw_term();

   alloc_bgr(HMC_BLOCKS,bs);
   error(assign_swd2swdbgr(HMC_BLOCKS,ODD_PTS)!=0,1,"main [time2b.c]",
         "Inversion of swd on the odd sites was not safe");

   wdt=1.0e6*wt_Qhat()/(double)(vol);     
   error_chk();

   if (my_rank==0)
   {
      printf("Time per lattice point: %4.3f micro sec",wdt);
      printf(" (%d Mflops [%d bit arithmetic])\n\n",
             (int)(1896.0/wdt),(int)(sizeof(spinor_dble))/3);
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
