
/*******************************************************************************
*
* File check10.c
*
* Copyright (C) 2005, 2007, 2008 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of hmc_gcr_vec(), hmc_gcr_frf() and hmc_gcr_acf()
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "block.h"
#include "linalg.h"
#include "dirac.h"
#include "sw_term.h"
#include "hmc.h"
#include "dfl.h"
#include "global.h"

static lat_parms_t lat;
static gcr_parms_t gcr;
static hmc_parms_t hmc;
static dfl_parms_t dfl;


static void init_mom(int n)
{
   hmc_t *hflds;

   hflds=hmcflds()+n;
   random_alg((*hflds).nlks,(*hflds).mom);
}


static void rot_ud(int n,double t)
{
   int nb,isw,*lk,*lkm;
   su3_dble *ub;
   su3_alg_dble *mom;
   block_t *b;
   hmc_t *hflds;

   b=blk_list(HMC_BLOCKS,&nb,&isw)+n;
   hflds=hmcflds()+n;
   
   ub=(*b).ud[0];
   mom=(*hflds).mom;
   lk=(*hflds).lks;
   lkm=lk+(*hflds).nlks;
   
   for (;lk<lkm;lk++)
   {
      ExpX(t,mom,ub+(*lk));
      mom+=1;
   }

   set_blk_flags(HMC_BLOCKS,NEW_ACTIVE_UD,n);
}


static void assign_block_action(double *act,int k)
{
   int nb,isw;
   block_t *b,*bm;

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   bm=b+nb;

   for (;b<bm;b++)
   {
      assign_sd2sdb(b,ALL_PTS,k,0);
      (*act)=norm_square_dble((*b).vol,0,(*b).sd[0]);
      
      act+=1;
   }
}


static double sum_action(hmc_act_t act)
{
   int nb,n;
   double *a0,*a1,*a2,sm,sall;

   nb=act.nb;
   a0=act.act0;
   a1=act.act1;
   a2=act.act2;   

   sm=0.0;

   for (n=0;n<nb;n++)
      sm+=(a0[n]+a1[n]+a2[n]);

   if (NPROC>1)
   {
      MPI_Reduce(&sm,&sall,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Bcast(&sall,1,MPI_DOUBLE,0,MPI_COMM_WORLD);     
   }
   else
      sall=sm;

   return sall;
}


static double dSdt(int n)
{
   double sm;
   su3_alg_dble *fr,*frm,*mom;
   hmc_t *hflds;

   hflds=hmcflds()+n;

   fr=(*hflds).frc;
   frm=fr+(*hflds).nlks;
   mom=(*hflds).mom;   
   sm=0.0;
   
   for (;fr<frm;fr++)
   {
      sm+=12.0*((*mom).c1*(*fr).c1+(*mom).c2*(*fr).c2);
      sm-= 6.0*((*mom).c1*(*fr).c2+(*mom).c2*(*fr).c1);
      sm+= 4.0*((*mom).c3*(*fr).c3+(*mom).c4*(*fr).c4+(*mom).c5*(*fr).c5+
                (*mom).c6*(*fr).c6+(*mom).c7*(*fr).c7+(*mom).c8*(*fr).c8);
      mom+=1;
   }

   return sm;
}


int main(int argc,char *argv[])
{
   int my_rank,nodfl,id;
   int bc,nc_first,nc_last,nc_step;
   int bs_hmc[4],N0,N1,N2;
   int nmx_acf,nmx_gcr_acf,nmx_rop,nmx_frf,nmx_gcr_frf;
   int bs_sap[4],nmr_sap,ncy_sap,nkv_sap;
   int bs_dfl[4],Ns,rinv,rnmr,rncy,rvnmx;
   int vnkv,vnmx,vdnmx;
   int n,i,isw,nsize,icnfg,i0,i1,i2,i3;
   int dfl_stat,status[3],ifail;
   int stat_min,stat_max,dmy;
   double beta,kappa,csw;
   double tau,res_acf,res_gcr_acf,res_rop,res_frf,res_gcr_frf;
   double rkappa,rvres,vres,vdres;
   double eps,dsdt,sact0,sact1,dact;
   double dev_frc,sig_loss;
   complex_dble z;
   spinor_dble **wsd;
   char cnfg_dir[NAME_SIZE],cnfg_file[NAME_SIZE];
   char nbase[NAME_SIZE];   
   hmc_act_t act0,act1,act2,act3;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check10.log","w",stdout);
      fin=freopen("check10.in","r",stdin);

      nodfl=find_opt(argc,argv,"-nodfl");
      
      printf("\n");
      printf("Check of hmc_gcr_vec(), hmc_gcr_frf() and hmc_gcr_acf()\n");
      printf("-------------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("id","%d\n",&id);
      read_line("cnfg_dir","%s\n",cnfg_dir);
      read_line("beta","%lf\n",&beta);      
      read_line("kappa","%lf\n",&kappa);
      read_line("csw","%lf\n",&csw);
      read_line("bc","%d\n",&bc);
      read_line("nc_first","%d\n",&nc_first);
      read_line("nc_last","%d\n",&nc_last);      
      read_line("nc_step","%d\n",&nc_step);
      
      read_line("bs_hmc","%d %d %d %d\n",
                &bs_hmc[0],&bs_hmc[1],&bs_hmc[2],&bs_hmc[3]);
      read_line("nmx_rop","%d\n",&nmx_rop);
      read_line("res_rop","%lf\n",&res_rop);
      read_line("nmx_gcr_acf","%d\n",&nmx_gcr_acf);
      read_line("res_gcr_acf","%lf\n",&res_gcr_acf);
      read_line("nmx_gcr_frf","%d\n",&nmx_gcr_frf);
      read_line("res_gcr_frf","%lf\n",&res_gcr_frf);       

      read_line("bs_sap","%d %d %d %d\n",
                &bs_sap[0],&bs_sap[1],&bs_sap[2],&bs_sap[3]);
      read_line("nmr_sap","%d\n",&nmr_sap);
      read_line("ncy_sap","%d\n",&ncy_sap);
      read_line("nkv_sap","%d\n",&nkv_sap);      

      if (nodfl==0)
      {
         read_line("bs_dfl","%d %d %d %d\n",
                   &bs_dfl[0],&bs_dfl[1],&bs_dfl[2],&bs_dfl[3]);
         read_line("Ns","%d\n",&Ns);
         read_line("rkappa","%lf\n",&rkappa);
         read_line("rinv","%d\n",&rinv);
         read_line("rnmr","%d\n",&rnmr);
         read_line("rncy","%d\n",&rncy);
         read_line("rvnmx","%d\n",&rvnmx);      
         read_line("rvres","%lf\n",&rvres);
         read_line("vnkv","%d\n",&vnkv);
         read_line("vnmx","%d\n",&vnmx);
         read_line("vres","%lf\n",&vres);
         read_line("vdnmx","%d\n",&vdnmx);
         read_line("vdres","%lf",&vdres);
      }
      
      fclose(fin);
   }

   MPI_Bcast(&nodfl,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&id,1,MPI_INT,0,MPI_COMM_WORLD);      
   MPI_Bcast(cnfg_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(&beta,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&kappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&csw,1,MPI_DOUBLE,0,MPI_COMM_WORLD);   
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_first,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_last,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&nc_step,1,MPI_INT,0,MPI_COMM_WORLD);   

   MPI_Bcast(bs_hmc,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_rop,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_rop,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_gcr_acf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_gcr_acf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_gcr_frf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_gcr_frf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

   MPI_Bcast(bs_sap,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmr_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncy_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nkv_sap,1,MPI_INT,0,MPI_COMM_WORLD);   

   if (nodfl==0)
   {
      MPI_Bcast(bs_dfl,4,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(&Ns,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rkappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&rinv,1,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(&rnmr,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rncy,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rvnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rvres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&vnkv,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&vnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&vres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&vdnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&vdres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }   
   
   start_ranlux(0,1234);
   geometry();
   alloc_ud();

   if (nodfl==0)
   {
      alloc_sd(11);
      
      if (Ns<(2*nkv_sap))
         alloc_s(2*nkv_sap+1);
      else
         alloc_s(Ns+2);

      alloc_u();
      alloc_sw();
      dfl=set_dfl_parms(bs_dfl,Ns,rinv,rnmr,rncy,rvnmx,vnkv,vnmx,vdnmx,
                        rkappa,rvres,vres,vdres,0,0);
   }
   else
   {
      alloc_sd(9);
      alloc_s(2*nkv_sap+1);
   }
   
   N0=1;
   N1=1;
   N2=1;
   nmx_acf=1;
   nmx_frf=1;
   tau=0.0;
   res_acf=1.0;
   res_frf=1.0;

   lat=set_lat_parms(beta,kappa,csw);
   gcr=set_gcr_parms(bs_sap,nmr_sap,ncy_sap,nkv_sap);
   hmc=set_hmc_parms(bs_hmc,bc,N0,N1,N2,nodfl==0,3,2,0,
                     nmx_acf,nmx_gcr_acf,nmx_rop,nmx_frf,nmx_gcr_frf,
                     tau,
                     res_acf,res_gcr_acf,res_rop,res_frf,res_gcr_frf);

   if (my_rank==0)
   {
      printf("beta = %.6f\n",lat.beta);
      printf("kappa = %.6f\n",lat.kappa);      
      printf("csw = %.6f\n",lat.csw);
      printf("bc = %d\n\n",hmc.bc);            

      printf("bs_hmc = %d %d %d %d\n",bs_hmc[0],bs_hmc[1],bs_hmc[2],bs_hmc[3]);
      printf("nmx_rop = %d\n",hmc.nmx_rop);            
      printf("res_rop = %.2e\n",hmc.res_rop);      
      printf("nmx_gcr_acf = %d\n",hmc.nmx_gcr_acf);            
      printf("res_gcr_acf = %.2e\n",hmc.res_gcr_acf);
      printf("nmx_gcr_frf = %d\n",hmc.nmx_gcr_frf);            
      printf("res_gcr_frf = %.2e\n\n",hmc.res_gcr_frf);

      printf("bs_sap = %d %d %d %d\n",bs_sap[0],bs_sap[1],bs_sap[2],bs_sap[3]);
      printf("nmr_sap = %d\n",gcr.nmr);                  
      printf("ncy_sap = %d\n",gcr.ncy);
      printf("nkv_sap = %d\n\n",gcr.nkv);  

      if (nodfl==0)
      {
         printf("bs_dfl = %d %d %d %d\n",
                bs_dfl[0],bs_dfl[1],bs_dfl[2],bs_dfl[3]);
         printf("Ns = %d\n",dfl.Ns);
         printf("rkappa = %.6f\n",dfl.rkappa);
         printf("rinv = %d\n",dfl.rinv);
         printf("rnmr = %d\n",dfl.rnmr);
         printf("rncy = %d\n",dfl.rncy);         
         printf("rvnmx = %d\n",dfl.rvnmx);
         printf("rvres = %.1e\n",dfl.rvres);
         printf("vnkv = %d\n",dfl.vnkv);         
         printf("vnmx = %d\n",dfl.vnmx);
         printf("vres = %.1e\n",dfl.vres);
         printf("vdnmx = %d\n",dfl.vdnmx);
         printf("vdres = %.1e\n\n",dfl.vdres);
      }

      printf("Configurations no %d -> %d in steps of %d\n\n",
             nc_first,nc_last,nc_step);
      fflush(flog);
   }

   error_root(((nc_last-nc_first)%nc_step)!=0,1,"main [check10.c]",
              "nc_last-nc_first is not a multiple of nc_step");
   check_dir_root(cnfg_dir);   

   nsize=name_size("%dx%dx%dx%db%.2fk%.5fc%.5fid%d",
                   NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);
   error_root(nsize>=NAME_SIZE,1,"main [check10.c]","Base name is too long");
   sprintf(nbase,"%dx%dx%dx%db%.2fk%.5fc%.5fid%d",
           NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);

   nsize=name_size("%s/%sn%d",cnfg_dir,nbase,nc_last);
   error_root(nsize>=NAME_SIZE,1,"main [check10.c]",
              "cnfg_dir name is too long");

   alloc_hmc_action(&act0);
   alloc_hmc_action(&act1);   
   alloc_hmc_action(&act2);
   alloc_hmc_action(&act3);
   
   wsd=reserve_wsd(4);
   i0=(wsd[0]-psd[0][0])/NSPIN;
   i1=(wsd[1]-psd[0][0])/NSPIN;
   i2=(wsd[2]-psd[0][0])/NSPIN;
   i3=(wsd[3]-psd[0][0])/NSPIN;

   for (icnfg=nc_first;icnfg<=nc_last;icnfg+=nc_step)
   {
      if (my_rank==0)
      {
         printf("Configuration no %d:\n\n",icnfg);
         fflush(flog);
      }

      sprintf(cnfg_file,"%s/%sn%d",cnfg_dir,nbase,icnfg);
      import_cnfg(cnfg_file);   
      if (bc==-1)
         flipbcd(); 
      assign_ud2udbgr(HMC_BLOCKS);
      swb_term();

      if (nodfl==0)
      {
         dfl_modes(0,&dfl_stat,&ifail);

         if (my_rank==0)
            printf("Generation of the deflation subspace: dfl_stat = %d, "
                   "ifail = %d\n",dfl_stat,ifail);

         error((dfl_stat<0)||(ifail!=0),1,"main [check10.c]",
               "dfl_modes() failed");
      }
         
      for (n=0;n<hmc.nb;n++)
      {
         init_mom(n);
         rot_ud(n,0.02);
      }

      reset_chrono();
      random_chi();
      assign_chi2sd(i0);
      assign_block_action(act0.act2,i0);      
      sact0=norm_square_dble(VOLUME,1,wsd[0]);
      ROp(status,&ifail);

      if (my_rank==0)
      {
         printf("Initialization of the pseudo-fermion field\n");
         printf("Application of ROp(): status = %d, ifail = %d\n\n",
                status[0],ifail);
         fflush(flog);
      }

      for (i=0;i<5;i++)
      {
         if (i>0)
         {
            assign_block_action(act0.act2,i0);                           
            sact0=norm_square_dble(VOLUME,1,wsd[0]);
            eps=1.0e-3;
               
            for (n=0;n<hmc.nb;n++)
               rot_ud(n,eps);

            assign_sd2chi(i0);
            ROp(status,&ifail);
         }

         hmc_gcr_vec(i1,i2,(i<2),&dfl_stat,status,&ifail);

         assign_chi2sd(i3);
         sub_assign_PdOs(i3,i1);
         z.re=-1.0;
         z.im=0.0;
         mulc_spinor_add_dble(VOLUME,wsd[3],wsd[0],z);
         eps=norm_square_dble(VOLUME,1,wsd[3])/sact0;
      
         if (my_rank==0)
         {
            if (i<3)
               printf("nold0 = %d, ",i);
            else
               printf("nold0 = %d, ",3);

            if (i<2)
               printf("nold1 = %d\n",i);
            else
               printf("nold1 = %d\n",2);

            printf("Application of hmc_gcr_vec(): ");
            
            if (nodfl)
               printf("status = %d, ifail = %d\n",status[0],ifail);
            else
               printf("dfl_stat = %d, status = %d (little: %d, %d), "
                      "ifail = %d\n",
                      dfl_stat,status[0],status[1],status[2],ifail);
            printf("Check of first vector  = %.2e\n",sqrt(eps));
            fflush(flog);
         }

         swap_active(HMC_BLOCKS);
         sw_term();
         Qnohat_dble(i2,i3);
         swap_active(HMC_BLOCKS);
         mulc_spinor_add_dble(VOLUME,wsd[3],wsd[0],z);
         eps=norm_square_dble(VOLUME,1,wsd[3])/sact0;

         if (my_rank==0)
         {
            printf("Check of second vector = %.2e\n\n",sqrt(eps));
            fflush(flog);
         }

         hmc_gcr_acf(i1,0,act1.act2,&dmy,status,&ifail);
         sact1=sum_action(act1);

         dact=0.0;
      
         for (n=0;n<act0.nb;n++)
         {
            eps=fabs(act0.act2[n]-act1.act2[n]);
         
            if (eps>dact)
               dact=eps;
         }

         MPI_Reduce(&dact,&eps,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
         
         if (my_rank==0)
         {
            printf("hmc_gcr_acf() using good vector: ");

            if (nodfl)
               printf("status = %d, ifail = %d\n",status[0],ifail);
            else
               printf("status = %d (little: %d, %d), ifail = %d\n",
                      status[0],status[1],status[2],ifail);         
      
            printf("Relative deviation of the action = %.2e\n",
                   fabs(1.0-sact1/sact0));

            printf("Maximal deviation of the block contributions = %.2e\n",
                   fabs((double)(act0.nb*NPROC)*eps/sact0));
            fflush(flog);
         }

         hmc_gcr_acf(i2,0,act1.act2,&dmy,status,&ifail);
         sact1=sum_action(act1);
      
         if (my_rank==0)
         {
            printf("hmc_gcr_acf() using false vector: ");

            if (nodfl)
               printf("status = %d, ifail = %d\n",status[0],ifail);
            else
               printf("status = %d (little: %d, %d), ifail = %d\n",
                      status[0],status[1],status[2],ifail);         
      
            printf("Relative deviation of the action = %.2e\n",
                   fabs(1.0-sact1/sact0));
            fflush(flog);
         }
   
         eps=hmc.res_gcr_acf;
         hmc=set_hmc_parms(bs_hmc,bc,N0,N1,N2,nodfl==0,3,2,0,
                           nmx_acf,nmx_gcr_acf,nmx_rop,nmx_frf,nmx_gcr_frf,
                           tau,
                           res_acf,10.0*res_gcr_frf,res_rop,res_frf,
                           res_gcr_frf);
   
         hmc_gcr_acf(i1,0,act1.act2,&dmy,status,&ifail);
         sact1=sum_action(act1);

         hmc=set_hmc_parms(bs_hmc,bc,N0,N1,N2,nodfl==0,3,2,0,
                           nmx_acf,nmx_gcr_acf,nmx_rop,nmx_frf,nmx_gcr_frf,
                           tau,
                           res_acf,eps,res_rop,res_frf,res_gcr_frf);
         error_chk();

         if (my_rank==0)
         {
            printf("hmc_gcr_acf() with low precision: ");

            if (nodfl)
               printf("status = %d, ifail = %d\n",status[0],ifail);
            else
               printf("status = %d (little: %d,%d), ifail = %d\n",
                      status[0],status[1],status[2],ifail);         
      
            printf("Relative deviation of the action = %.2e\n\n",
                   fabs(1.0-sact1/sact0));
            fflush(flog);
         }

         dact=0.0;

         for (n=0;n<hmc.nb;n++)
         {
            hmc_gcr_frf(n,i1,i2);
            dact+=dSdt(n);
         }

         MPI_Reduce(&dact,&dsdt,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
         MPI_Bcast(&dsdt,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

         ifail=0;
         stat_max=0;
         stat_min=0;
         eps=1.0e-3;
   
         for (n=0;n<hmc.nb;n++)
            rot_ud(n,eps);

         hmc_gcr_acf(i1,0,act0.act2,&dmy,status,&isw);
         sact0=2.0*sum_action(act0)/3.0;

         for (n=0;n<hmc.nb;n++)
            rot_ud(n,-eps);

         ifail|=isw;
         if (status[0]>stat_max)
            stat_max=status[0];
         if ((stat_min==0)||(status[0]<stat_min))
            stat_min=status[0];

         for (n=0;n<hmc.nb;n++)
            rot_ud(n,-eps);

         hmc_gcr_acf(i1,0,act1.act2,&dmy,status,&isw);
         sact1=2.0*sum_action(act1)/3.0;
      
         for (n=0;n<hmc.nb;n++)
            rot_ud(n,eps);

         ifail|=isw;
         if (status[0]>stat_max)
            stat_max=status[0];
         if ((stat_min==0)||(status[0]<stat_min))
            stat_min=status[0];

         for (n=0;n<hmc.nb;n++)
            rot_ud(n,2.0*eps);

         hmc_gcr_acf(i1,0,act2.act2,&dmy,status,&isw);
         sact0-=sum_action(act2)/12.0;

         for (n=0;n<hmc.nb;n++)
            rot_ud(n,-2.0*eps);

         ifail|=isw;
         if (status[0]>stat_max)
            stat_max=status[0];
         if ((stat_min==0)||(status[0]<stat_min))
            stat_min=status[0];

         for (n=0;n<hmc.nb;n++)
            rot_ud(n,-2.0*eps);

         hmc_gcr_acf(i1,0,act3.act2,&dmy,status,&isw);
         sact1-=sum_action(act3)/12.0;

         for (n=0;n<hmc.nb;n++)
            rot_ud(n,2.0*eps);

         ifail|=isw;
         if (status[0]>stat_max)
            stat_max=status[0];
         if ((stat_min==0)||(status[0]<stat_min))
            stat_min=status[0];
      
         dact=(sact0-sact1)/eps;
         dev_frc=fabs(1.0-dsdt/dact);
         sig_loss=-log10(fabs(1.0-sact0/sact1));

         error_chk();
         MPI_Reduce(&stat_max,status,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
         stat_max=status[0];
         MPI_Reduce(&stat_min,status,1,MPI_INT,MPI_MIN,0,MPI_COMM_WORLD);
         stat_min=status[0];
         MPI_Reduce(&ifail,&isw,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
         ifail=isw;

         if (my_rank==0)
         {
            printf("Calculation of the force:\n");
            printf("Relative deviation of dS/dt = %.2e ",dev_frc);
            printf("[significance loss = %d digits]\n",(int)(sig_loss));
            printf("status             = %d ... %d, ifail = %d\n\n",
                   stat_min,stat_max,ifail);
            fflush(flog);
         }

         sact0=0.0;

         for (n=0;n<act0.nb;n++)
         {
            sact0+=2.0*(act0.act2[n]-act1.act2[n])/3.0;
            sact0-=(act2.act2[n]-act3.act2[n])/12.0;
         }

         MPI_Reduce(&sact0,&dact,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
         MPI_Bcast(&dact,1,MPI_DOUBLE,0,MPI_COMM_WORLD); 

         dact/=eps;
         dev_frc=fabs(1.0-dsdt/dact);      
      
         if (my_rank==0)
         {
            printf("Using blocked action for the numerical differentiation:\n");
            printf("Relative deviation of dS/dt = %.2e ",dev_frc);
            printf("[epsilon = %.1e]\n\n",eps);
            fflush(flog);
         }
      }
   }

   if (my_rank==0)
      fclose(flog);
   
   MPI_Finalize();    
   exit(0);
}
