
/*******************************************************************************
*
* File check11.c
*
* Copyright (C) 2007, 2008 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Numerical accuracy of the HMD action
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "block.h"
#include "sw_term.h"
#include "dirac.h"
#include "hmc.h"
#include "dfl.h"
#include "global.h"

static lat_parms_t lat;
static gcr_parms_t gcr;
static hmc_parms_t hmc;
static dfl_parms_t dfl;


static double cmp_ud(su3_dble *u,su3_dble *v)
{
   double *ru,*rv,*rm;
   double d,d1,d2,dmax;

   ru=(double*)(u);
   rv=(double*)(v);
   rm=ru+18;
   dmax=0.0;

   for (;ru<rm;ru+=2)
   {
      d1=(*(ru  ))-(*(rv  ));
      d2=(*(ru+1))-(*(rv+1));      
      d=d1*d1+d2*d2;

      if (d>dmax)
         dmax=d;
      
      rv+=2;
   }

   return dmax;
}


static double max_dev_ud(void)
{
   int nb,isw,ix,iy,mu;
   double d,dmax;
   su3_dble *u,*v;
   block_t *b,*bm;

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   bm=b+nb;
   dmax=0.0;
   
   for (;b<bm;b++)
   {
      u=(*b).ud[0];
         
      for (ix=(*b).volh;ix<(*b).vol;ix++)
      {
         iy=(*b).imb[ix];
         v=pud[iy][0];   
         
         for (mu=0;mu<4;mu++)
         {
            if ((*b).iup[ix][mu]!=(*b).vol)
            {
               d=cmp_ud(u,v);
               if (d>dmax)
                  dmax=d;
            }

            u+=1;
            v+=1;
         
            if ((*b).idn[ix][mu]!=(*b).vol)
            {
               d=cmp_ud(u,v);
               if (d>dmax)
                  dmax=d;
            }

            u+=1;
            v+=1;          
         }
      }
   }

   MPI_Reduce(&dmax,&d,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
   MPI_Bcast(&d,1,MPI_DOUBLE,0,MPI_COMM_WORLD);   
   
   return sqrt(d);
}


static void sum_action(hmc_act_t act,double *smax,double *sact)
{
   int n,nb;
   double d,sm[3];

   nb=act.nb;
   sm[0]=0.0;
   sm[1]=0.0;
   sm[2]=0.0;

   for (n=0;n<nb;n++)
   {
      d=fabs(act.act0[n]);
      if (d>sm[0])
         sm[0]=d;
      d=fabs(act.act1[n]);
      if (d>sm[1])
         sm[1]=d;
      d=fabs(act.act2[n]);
      if (d>sm[2])
         sm[2]=d;      
   }
   
   MPI_Reduce(sm,smax,3,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
   MPI_Bcast(smax,3,MPI_DOUBLE,0,MPI_COMM_WORLD);

   sm[0]=0.0;
   sm[1]=0.0;
   sm[2]=0.0;   
   
   for (n=0;n<nb;n++)
   {
      sm[0]+=act.act0[n];
      sm[1]+=act.act1[n];      
      sm[2]+=act.act2[n];
   }

   MPI_Reduce(sm,sact,3,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
   MPI_Bcast(sact,3,MPI_DOUBLE,0,MPI_COMM_WORLD);     
}


static void diff_action(hmc_act_t act0,hmc_act_t act1,double *dmax,double *dact)
{
   int n,nb;   
   double d,sm[3];

   nb=act0.nb;
   sm[0]=0.0;
   sm[1]=0.0;
   sm[2]=0.0;

   for (n=0;n<nb;n++)
   {
      d=fabs(act1.act0[n]-act0.act0[n]);
      if (d>sm[0])
         sm[0]=d;
      d=fabs(act1.act1[n]-act0.act1[n]);
      if (d>sm[1])
         sm[1]=d;
      d=fabs(act1.act2[n]-act0.act2[n]);
      if (d>sm[2])
         sm[2]=d;      
   }
   
   MPI_Reduce(sm,dmax,3,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
   MPI_Bcast(dmax,3,MPI_DOUBLE,0,MPI_COMM_WORLD);

   sm[0]=0;
   sm[1]=0;
   sm[2]=0;

   for (n=0;n<nb;n++)
   {
      sm[0]+=(act1.act0[n]-act0.act0[n]);
      sm[1]+=(act1.act1[n]-act0.act1[n]);
      sm[2]+=(act1.act2[n]-act0.act2[n]);
   }

   MPI_Reduce(sm,dact,3,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
   MPI_Bcast(dact,3,MPI_DOUBLE,0,MPI_COMM_WORLD);   
}


int main(int argc,char *argv[])
{
   int my_rank,nodfl,id;
   int bc,nc_first,nc_last,nc_step;
   int bs_hmc[4],nmx_acf,nmx_gcr_acf,nmx_rop;
   int bs_sap[4],nmr_sap,ncy_sap,nkv_sap;
   int bs_dfl[4],Ns,rinv,rnmr,rncy,rvnmx;
   int vnkv,vnmx,vdnmx;
   int stat_rop,stat_acf,stat_gcr_acf[3];
   int stat_dfl_mds,ifail;
   int n,nh,nsize,icnfg,iwd0,dmy;
   int stat,stat_min,stat_sum,flag;
   int stat_rop_min,stat_rop_max,stat_rop_avg;
   int stat_acf_min,stat_acf_max,stat_acf_avg;
   int stat_gcr_acf_min[3],stat_gcr_acf_max[3],stat_gcr_acf_avg[3];
   double beta,kappa,csw;
   double res_acf,res_gcr_acf,res_rop;
   double rkappa,rvres,vres,vdres;
   double dU,smax[3],sact[3],dmax[3],dact[3];
   double dB,dBmin,dBmax,dBavg;
   double dG,dGmin,dGmax,dGavg;   
   double dH,dHmin,dHmax,dHavg;
   spinor_dble **wsd;
   char cnfg_dir[NAME_SIZE],cnfg_file[NAME_SIZE];
   char nbase[NAME_SIZE];   
   hmc_act_t act0,act1;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check11.log","w",stdout);
      fin=freopen("check11.in","r",stdin);

      nodfl=find_opt(argc,argv,"-nodfl");      
      
      printf("\n");
      printf("Numerical accuracy of the HMD action\n");
      printf("------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("id","%d\n",&id);
      read_line("cnfg_dir","%s\n",cnfg_dir);
      read_line("beta","%lf\n",&beta);      
      read_line("kappa","%lf\n",&kappa);
      read_line("csw","%lf\n",&csw);
      read_line("bc","%d\n",&bc);
      read_line("nc_first","%d\n",&nc_first);
      read_line("nc_last","%d\n",&nc_last);  
      read_line("nc_step","%d\n",&nc_step);  
      
      read_line("bs_hmc","%d %d %d %d\n",
                &bs_hmc[0],&bs_hmc[1],&bs_hmc[2],&bs_hmc[3]);
      read_line("nmx_rop","%d\n",&nmx_rop);
      read_line("res_rop","%lf\n",&res_rop);
      read_line("nmx_acf","%d\n",&nmx_acf);
      read_line("res_acf","%lf\n",&res_acf);
      read_line("nmx_gcr_acf","%d\n",&nmx_gcr_acf);
      read_line("res_gcr_acf","%lf\n",&res_gcr_acf);

      read_line("bs_sap","%d %d %d %d\n",
                &bs_sap[0],&bs_sap[1],&bs_sap[2],&bs_sap[3]);
      read_line("nmr_sap","%d\n",&nmr_sap);
      read_line("ncy_sap","%d\n",&ncy_sap);
      read_line("nkv_sap","%d\n",&nkv_sap);      

      if (nodfl==0)
      {
         read_line("bs_dfl","%d %d %d %d\n",
                   &bs_dfl[0],&bs_dfl[1],&bs_dfl[2],&bs_dfl[3]);
         read_line("Ns","%d\n",&Ns);
         read_line("rkappa","%lf\n",&rkappa);
         read_line("rinv","%d\n",&rinv);
         read_line("rnmr","%d\n",&rnmr);
         read_line("rncy","%d\n",&rncy);
         read_line("rvnmx","%d\n",&rvnmx);      
         read_line("rvres","%lf\n",&rvres);
         read_line("vnkv","%d\n",&vnkv);
         read_line("vnmx","%d\n",&vnmx);
         read_line("vres","%lf\n",&vres);
         read_line("vdnmx","%d\n",&vdnmx);
         read_line("vdres","%lf",&vdres);
      }

      fclose(fin);
   }

   MPI_Bcast(&nodfl,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&id,1,MPI_INT,0,MPI_COMM_WORLD);      
   MPI_Bcast(cnfg_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(&beta,1,MPI_DOUBLE,0,MPI_COMM_WORLD);   
   MPI_Bcast(&kappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&csw,1,MPI_DOUBLE,0,MPI_COMM_WORLD);   
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_first,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_last,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_step,1,MPI_INT,0,MPI_COMM_WORLD);
   
   MPI_Bcast(bs_hmc,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_rop,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_rop,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_acf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_acf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_gcr_acf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_gcr_acf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

   MPI_Bcast(bs_sap,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmr_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncy_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nkv_sap,1,MPI_INT,0,MPI_COMM_WORLD);   
   
   if (nodfl==0)
   {
      MPI_Bcast(bs_dfl,4,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(&Ns,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rkappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&rinv,1,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(&rnmr,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rncy,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rvnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rvres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&vnkv,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&vnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&vres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&vdnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&vdres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }   
   
   start_ranlux(0,1234);
   geometry();
   alloc_ud();

   if (nodfl==0)
   {
      alloc_sd(7);
      
      if (Ns<(2*nkv_sap))
         alloc_s(2*nkv_sap+1);
      else
         alloc_s(Ns+2);

      alloc_u();
      alloc_sw();
      dfl=set_dfl_parms(bs_dfl,Ns,rinv,rnmr,rncy,rvnmx,vnkv,vnmx,vdnmx,
                        rkappa,rvres,vres,vdres,0,0);
   }
   else
   {
      alloc_sd(4);
      alloc_s(2*nkv_sap+1);
   }   

   lat=set_lat_parms(beta,kappa,csw);
   gcr=set_gcr_parms(bs_sap,nmr_sap,ncy_sap,nkv_sap);
   hmc=set_hmc_parms(bs_hmc,bc,1,1,1,nodfl==0,0,0,0,
                     nmx_acf,nmx_gcr_acf,nmx_rop,1,1,
                     0.5,
                     res_acf,res_gcr_acf,res_rop,1.0,1.0);

   if (my_rank==0)
   {
      printf("beta = %.6f\n",lat.beta);      
      printf("kappa = %.6f\n",lat.kappa);      
      printf("csw = %.6f\n",lat.csw);
      printf("bc = %d\n\n",hmc.bc);            

      printf("bs_hmc = %d %d %d %d\n",bs_hmc[0],bs_hmc[1],bs_hmc[2],bs_hmc[3]);
      printf("nmx_acf = %d\n",hmc.nmx_acf);            
      printf("res_acf = %.2e\n",hmc.res_acf);      
      printf("nmx_rop = %d\n",hmc.nmx_rop);            
      printf("res_rop = %.2e\n",hmc.res_rop);      
      printf("nmx_gcr_acf = %d\n",hmc.nmx_gcr_acf);            
      printf("res_gcr_acf = %.2e\n\n",hmc.res_gcr_acf);
      
      printf("bs_sap = %d %d %d %d\n",bs_sap[0],bs_sap[1],bs_sap[2],bs_sap[3]);
      printf("nmr = %d\n",gcr.nmr);                  
      printf("ncy = %d\n",gcr.ncy);
      printf("nkv = %d\n\n",gcr.nkv);                        

      if (nodfl==0)
      {
         printf("bs_dfl = %d %d %d %d\n",
                bs_dfl[0],bs_dfl[1],bs_dfl[2],bs_dfl[3]);
         printf("Ns = %d\n",dfl.Ns);
         printf("rkappa = %.6f\n",dfl.rkappa);
         printf("rinv = %d\n",dfl.rinv);
         printf("rnmr = %d\n",dfl.rnmr);
         printf("rncy = %d\n",dfl.rncy);         
         printf("rvnmx = %d\n",dfl.rvnmx);
         printf("rvres = %.1e\n",dfl.rvres);
         printf("vnkv = %d\n",dfl.vnkv);         
         printf("vnmx = %d\n",dfl.vnmx);
         printf("vres = %.1e\n",dfl.vres);
         printf("vdnmx = %d\n",dfl.vdnmx);
         printf("vdres = %.1e\n\n",dfl.vdres);
      }

      printf("Configurations no %d -> %d in steps of %d\n\n",
             nc_first,nc_last,nc_step);
      fflush(flog);
   }

   error_root(((nc_last-nc_first)%nc_step)!=0,1,"main [check11.c]",
              "nc_last-nc_first is not a multiple of nc_step");
   check_dir_root(cnfg_dir);   

   nsize=name_size("%dx%dx%dx%db%.2fk%.5fc%.5fid%d",
                   NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);
   error_root(nsize>=NAME_SIZE,1,"main [check11.c]","Base name is too long");
   sprintf(nbase,"%dx%dx%dx%db%.2fk%.5fc%.5fid%d",
           NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);

   nsize=name_size("%s/%sn%d",cnfg_dir,nbase,nc_last);
   error_root(nsize>=NAME_SIZE,1,"main [check11.c]",
              "cnfg_dir name is too long");

   alloc_hmc_action(&act0);
   alloc_hmc_action(&act1);

   wsd=reserve_wsd(1);
   iwd0=(wsd[0]-psd[0][0])/NSPIN;

   stat_rop_min=0;
   stat_rop_max=0;
   stat_rop_avg=0;
   
   stat_acf_min=0;
   stat_acf_max=0;
   stat_acf_avg=0;

   for (n=0;n<3;n++)
   {
      stat_gcr_acf_min[n]=0;
      stat_gcr_acf_max[n]=0;
      stat_gcr_acf_avg[n]=0;   
   }

   dBmin=0.0;
   dBmax=0.0;
   dBavg=0.0;

   dGmin=0.0;
   dGmax=0.0;
   dGavg=0.0;

   dHmin=0.0;
   dHmax=0.0;
   dHavg=0.0;
   
   for (icnfg=nc_first;icnfg<=nc_last;icnfg+=nc_step)
   {
      if (my_rank==0)
      {
         printf("Configuration no %d:\n\n",icnfg);
         fflush(flog);
      }   

      sprintf(cnfg_file,"%s/%sn%d",cnfg_dir,nbase,icnfg);
      import_cnfg(cnfg_file);  

      start_md(act0,&stat_rop,&stat_dfl_mds,&ifail);
      sum_action(act0,smax,sact);
      dU=max_dev_ud();
      error_root(dU!=0.0,1,"main [check11.c]","Gauge field on the "
                 "HMC_BLOCKS grid differs from the global one");

      if (my_rank==0)
      {
         printf("start_md:\n");
         if (nodfl)
            printf("stat_rop = %d, ifail = %d\n",stat_rop,ifail);
         else
            printf("stat_rop = %d, stat_dfl_mds = %d, ifail = %d\n",
                   stat_rop,stat_dfl_mds,ifail);
         printf("Maximal action part "
                "(gauge, block, global) = %.2e, %.2e, %.2e\n",
                smax[0],smax[1],smax[2]);
         printf("Total action (gauge, block, global) = %.2e, %.2e, %.2e\n",
                sact[0],sact[1],sact[2]);
         fflush(flog);
      }

      error_root((stat_rop<0)||(stat_dfl_mds<0)||(ifail!=0),1,
                 "main [check11.c]","start_md() failed");
      
      set_sd2zero(VOLUME,wsd[0]);
      hmc_gcr_acf(iwd0,0,act1.act2,&dmy,stat_gcr_acf,&ifail);

      if (my_rank==0)
      {
         printf("hmc_gcr_acf:\n");

         if (nodfl)
            printf("stat_gcr_acf = %d, ifail = %d\n",
                   stat_gcr_acf[0],ifail);
         else
            printf("stat_gcr_acf = %d (little: %d, %d), ifail = %d\n",
                   stat_gcr_acf[0],stat_gcr_acf[1],stat_gcr_acf[2],ifail);
      }

      error_root((stat_gcr_acf[0]<0)||(stat_gcr_acf[2]<0)||(ifail!=0),1,
                 "main [check11.c]","hmc_gcr_acf() failed");
      
      stat=0;
      flag=0;
      
      for (n=0;n<hmc.nb;n++)
      {
         act1.act1[n]=hmc_acf(n,1,1,&stat_acf,&flag);

         if (stat_acf>=0)
            stat+=stat_acf;
         else
            stat=stat_acf;

         if ((stat<0)||(flag!=0))
            break;         
      }
      
      MPI_Reduce(&stat,&stat_min,1,MPI_INT,MPI_MIN,0,MPI_COMM_WORLD);
      MPI_Reduce(&stat,&stat_sum,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Reduce(&flag,&ifail,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
      MPI_Bcast(&stat_min,1,MPI_INT,0,MPI_COMM_WORLD);      
      MPI_Bcast(&stat_sum,1,MPI_INT,0,MPI_COMM_WORLD);            
      MPI_Bcast(&ifail,1,MPI_INT,0,MPI_COMM_WORLD);
      
      if (stat_min<0)
         stat_acf=stat_min;
      else
      {
         n=NPROC*hmc.nb;
         nh=n/2;
         stat_acf=(stat_sum+nh)/n;
      }

      if (my_rank==0)
      {
         printf("hmc_acf:\n");
         printf("stat_acf = %d, ifail = %d\n",stat_acf,ifail);
         fflush(flog);
      }

      error_root((stat_acf<0)||(ifail!=0),1,"main [check11.c]",
                 "hmc_acf() failed");
      error_chk();

      diff_action(act0,act1,dmax,dact);
      dB=dmax[1];
      dG=dmax[2];
      dH=fabs(dact[1]+dact[2]);

      if (my_rank==0)
      {
         printf("Difference of action parts "
                "(block, global) <= %.2e, %.2e\n",dB,dG);
         printf("|Delta H| = %.2e\n\n",dH);
         fflush(flog);
      }
      
      if (icnfg==nc_first)
      {
         stat_rop_min=stat_rop;
         stat_rop_max=stat_rop;
         stat_rop_avg=stat_rop;         
         
         stat_acf_min=stat_acf;
         stat_acf_max=stat_acf;
         stat_acf_avg=stat_acf;

         for (n=0;n<3;n++)
         {
            stat_gcr_acf_min[n]=stat_gcr_acf[n];
            stat_gcr_acf_max[n]=stat_gcr_acf[n];
            stat_gcr_acf_avg[n]=stat_gcr_acf[n];
         }
      }
      else
      {
         if (stat_rop<stat_rop_min)
            stat_rop_min=stat_rop;
         if (stat_rop>stat_rop_max)
            stat_rop_max=stat_rop;
         stat_rop_avg+=stat_rop;
         
         if (stat_acf<stat_acf_min)
            stat_acf_min=stat_acf;
         if (stat_acf>stat_acf_max)
            stat_acf_max=stat_acf;
         stat_acf_avg+=stat_acf;

         for (n=0;n<3;n++)
         {
            if (stat_gcr_acf[n]<stat_gcr_acf_min[n])
               stat_gcr_acf_min[n]=stat_gcr_acf[n];
            if (stat_gcr_acf[n]>stat_gcr_acf_max[n])
               stat_gcr_acf_max[n]=stat_gcr_acf[n];
            stat_gcr_acf_avg[n]+=stat_gcr_acf[n];
         }
      }

      if (icnfg==nc_first)
      {
         dBmin=dB;
         dBmax=dB;
         dBavg=dB;

         dGmin=dG;
         dGmax=dG;
         dGavg=dG;         

         dHmin=dH;
         dHmax=dH;
         dHavg=dH;
      }
      else
      {
         if (dB<dBmin)
            dBmin=dB;
         if (dB>dBmax)
            dBmax=dB;
         dBavg+=dB;

         if (dG<dGmin)
            dGmin=dG;
         if (dG>dGmax)
            dGmax=dG;
         dGavg+=dG;         
         
         if (dH<dHmin)
            dHmin=dH;
         if (dH>dHmax)
            dHmax=dH;
         dHavg+=dH;
      }
   }

   if (my_rank==0)
   {
      n=(nc_last-nc_first)/nc_step+1;

      printf("Test summary\n");
      printf("------------\n\n");

      printf("Considered %d configurations in the range %d -> %d\n",
             n,nc_first,nc_last);      
      
      printf("The three figures quoted in each case are the minimal,\n");
      printf("maximal and average values\n\n");

      printf("stat_rop = %d, %d, %d\n",
             stat_rop_min,stat_rop_max,stat_rop_avg/n);
      printf("stat_acf = %d, %d, %d\n",
             stat_acf_min,stat_acf_max,stat_acf_avg/n);      

      if (nodfl)
         printf("stat_gcr_acf = %d, %d, %d\n\n",
                stat_gcr_acf_min[0],stat_gcr_acf_max[0],stat_gcr_acf_avg[0]/n);
      else
      {
         printf("stat_gcr_acf[0] = %d, %d, %d\n",
                stat_gcr_acf_min[0],stat_gcr_acf_max[0],stat_gcr_acf_avg[0]/n);
         printf("stat_gcr_acf[1] = %d, %d, %d\n",
                stat_gcr_acf_min[1],stat_gcr_acf_max[1],stat_gcr_acf_avg[1]/n);
         printf("stat_gcr_acf[2] = %d, %d, %d\n\n",
                stat_gcr_acf_min[2],stat_gcr_acf_max[2],stat_gcr_acf_avg[2]/n);
      }

      printf("max|Delta H_block| = %.2e, %.2e, % .2e\n",
             dBmin,dBmax,dBavg/(double)(n));
      printf("max|Delta H_global| = %.2e, %.2e, % .2e\n",
             dGmin,dGmax,dGavg/(double)(n));      
      printf("|Delta H| = %.2e, %.2e, % .2e\n\n",
             dHmin,dHmax,dHavg/(double)(n));

      fclose(flog);
   }
   
   MPI_Finalize();    
   exit(0);
}
