
/*******************************************************************************
*
* File check13.c
*
* Copyright (C) 2005, 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Conservation of the action by the HMD evolution
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "block.h"
#include "sw_term.h"
#include "dirac.h"
#include "hmc.h"
#include "dfl.h"
#include "global.h"

static su3_alg_dble *msv=NULL;
static lat_parms_t lat;
static gcr_parms_t gcr;
static hmc_parms_t hmc;
static dfl_parms_t dfl;


static void save_mom(void)
{
   int nb,isw;
   su3_alg_dble *mom,*momm,*momsv;
   hmc_t *hflds,*hfldsm;

   blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();
   hfldsm=hflds+nb;

   if (msv==NULL)
   {
      msv=amalloc(nb*(*hflds).nlks*sizeof(su3_alg_dble),ALIGN);
      error(msv==NULL,1,"save_mom [check13.c]",
            "Unable to allocate auxiliary array");
   }

   momsv=msv;
   
   for (;hflds<hfldsm;hflds++)
   {
      mom=(*hflds).mom;
      momm=mom+(*hflds).nlks;

      for (;mom<momm;mom++)
      {
         *momsv=*mom;
         momsv+=1;
      }
   }
}


static void restore_mom(void)
{
   int nb,isw;
   su3_alg_dble *mom,*momm,*momsv;
   hmc_t *hflds,*hfldsm;

   blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();
   hfldsm=hflds+nb;

   error(msv==NULL,1,"restore_mom [check13.c]",
         "The momenta must first be saved");
   momsv=msv;
   
   for (;hflds<hfldsm;hflds++)
   {
      mom=(*hflds).mom;
      momm=mom+(*hflds).nlks;

      for (;mom<momm;mom++)
      {
         *mom=*momsv;
         momsv+=1;
      }
   }
}


static void restart_md(int N0,int N1,int N2)
{
   int n,nb,isw,stat,flag;
   block_t *b;

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   hmc=set_hmc_parms((*b).bs,hmc.bc,N0,N1,N2,hmc.dfl,
                     hmc.cno0,hmc.cno1,hmc.cno_blk,
                     hmc.nmx_acf,hmc.nmx_gcr_acf,hmc.nmx_rop,
                     hmc.nmx_frf,hmc.nmx_gcr_frf,
                     hmc.tau,
                     hmc.res_acf,hmc.res_gcr_acf,hmc.res_rop,
                     hmc.res_frf,hmc.res_gcr_frf);
   
   n=query_flags(BCD_FLIPPED);
   if (((n!=0)&&(hmc.bc==1))||((n!=1)&&(hmc.bc==-1)))
      flipbcd();

   restore_mom();
   assign_ud2udbgr(HMC_BLOCKS);
   swb_term();

   if (hmc.dfl==1)
      dfl_modes(0,&stat,&flag);
}


static double sum_action(hmc_act_t act)
{
   int nb,n;
   double *a0,*a1,*a2,sm,sall;

   nb=act.nb;
   a0=act.act0;
   a1=act.act1;
   a2=act.act2;   

   sm=0.0;

   for (n=0;n<nb;n++)
      sm+=(a0[n]+a1[n]+a2[n]);

   if (NPROC>1)
   {
      MPI_Reduce(&sm,&sall,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Bcast(&sall,1,MPI_DOUBLE,0,MPI_COMM_WORLD);     
   }
   else
      sall=sm;

   return sall;
}


static double diff_action(hmc_act_t act0,hmc_act_t act1)
{
   int nb,n;
   double a0,a1,a2,sm,sall;

   nb=act0.nb;
   a0=0.0;
   a1=0.0;
   a2=0.0;

   for (n=0;n<nb;n++)
   {
      a0+=(act1.act0[n]-act0.act0[n]);
      a1+=(act1.act1[n]-act0.act1[n]);
      a2+=(act1.act2[n]-act0.act2[n]);
   }

   sm=a0+a1+a2;

   if (NPROC>1)
   {
      MPI_Reduce(&sm,&sall,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Bcast(&sall,1,MPI_DOUBLE,0,MPI_COMM_WORLD);     
   }
   else
      sall=sm;
   
   return sall;
}


int main(int argc,char *argv[])
{
   int my_rank,nodfl,id;
   int bc,nc_first,nc_last,nc_step;   
   int bs_hmc[4],N0,N1,N2[3],cno0,cno1,cno_blk;
   int nmx_acf,nmx_gcr_acf,nmx_rop,nmx_frf,nmx_gcr_frf;
   int bs_sap[4],nmr_sap,ncy_sap,nkv_sap;
   int bs_dfl[4],Ns,rinv,rnmr,rncy,rvnmx;
   int vnkv,vnmx,vdnmx;
   int stat_acf,stat_gcr_acf[3];
   int stat_rop,stat_frf,stat_gcr_frf[3],ifail;
   int stat_dfl_mds,stat_dfl_upd[2];
   int i,nsize,icnfg;
   double beta,kappa,csw;
   double tau,res_acf,res_gcr_acf,res_rop,res_frf,res_gcr_frf;
   double rkappa,rvres,vres,vdres;   
   double sact0,sact1,dact[3],eps0[3];
   char cnfg_dir[NAME_SIZE],cnfg_file[NAME_SIZE];
   char nbase[NAME_SIZE];      
   hmc_act_t act0,act1;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check13.log","w",stdout);
      fin=freopen("check13.in","r",stdin);

      nodfl=find_opt(argc,argv,"-nodfl");
      
      printf("\n");
      printf("Conservation of the action by the HMD evolution\n");
      printf("-----------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("id","%d\n",&id);
      read_line("cnfg_dir","%s\n",cnfg_dir);      
      read_line("beta","%lf\n",&beta);
      read_line("kappa","%lf\n",&kappa);
      read_line("csw","%lf\n",&csw);
      read_line("bc","%d\n",&bc);
      read_line("nc_first","%d\n",&nc_first);
      read_line("nc_last","%d\n",&nc_last);
      read_line("nc_step","%d\n",&nc_step);

      read_line("bs_hmc","%d %d %d %d\n",
                &bs_hmc[0],&bs_hmc[1],&bs_hmc[2],&bs_hmc[3]);
      read_line("tau","%lf\n",&tau);      
      read_line("N0","%d\n",&N0);
      read_line("N1","%d\n",&N1);
      read_line("N2","%d %d %d\n",&N2[0],&N2[1],&N2[2]); 
      read_line("cno0","%d\n",&cno0);
      read_line("cno1","%d\n",&cno1);
      read_line("cno_blk","%d\n",&cno_blk);      
      read_line("nmx_rop","%d\n",&nmx_rop);
      read_line("res_rop","%lf\n",&res_rop);
      read_line("nmx_acf","%d\n",&nmx_acf);
      read_line("res_acf","%lf\n",&res_acf);
      read_line("nmx_frf","%d\n",&nmx_frf);
      read_line("res_frf","%lf\n",&res_frf);
      read_line("nmx_gcr_acf","%d\n",&nmx_gcr_acf);
      read_line("res_gcr_acf","%lf\n",&res_gcr_acf);
      read_line("nmx_gcr_frf","%d\n",&nmx_gcr_frf);
      read_line("res_gcr_frf","%lf\n",&res_gcr_frf);       

      read_line("bs_sap","%d %d %d %d\n",
                &bs_sap[0],&bs_sap[1],&bs_sap[2],&bs_sap[3]);
      read_line("nmr_sap","%d\n",&nmr_sap);
      read_line("ncy_sap","%d\n",&ncy_sap);
      read_line("nkv_sap","%d\n",&nkv_sap);      

      if (nodfl==0)
      {
         read_line("bs_dfl","%d %d %d %d\n",
                   &bs_dfl[0],&bs_dfl[1],&bs_dfl[2],&bs_dfl[3]);
         read_line("Ns","%d\n",&Ns);
         read_line("rkappa","%lf\n",&rkappa);
         read_line("rinv","%d\n",&rinv);
         read_line("rnmr","%d\n",&rnmr);
         read_line("rncy","%d\n",&rncy);
         read_line("rvnmx","%d\n",&rvnmx);      
         read_line("rvres","%lf\n",&rvres);
         read_line("vnkv","%d\n",&vnkv);
         read_line("vnmx","%d\n",&vnmx);
         read_line("vres","%lf\n",&vres);
         read_line("vdnmx","%d\n",&vdnmx);
         read_line("vdres","%lf",&vdres);
      }

      fclose(fin);
   }

   MPI_Bcast(&nodfl,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&id,1,MPI_INT,0,MPI_COMM_WORLD);      
   MPI_Bcast(cnfg_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(&beta,1,MPI_DOUBLE,0,MPI_COMM_WORLD);   
   MPI_Bcast(&kappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&csw,1,MPI_DOUBLE,0,MPI_COMM_WORLD);   
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_first,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_last,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_step,1,MPI_INT,0,MPI_COMM_WORLD);
   
   MPI_Bcast(bs_hmc,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&tau,1,MPI_DOUBLE,0,MPI_COMM_WORLD);   
   MPI_Bcast(&N0,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&N1,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(N2,3,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&cno0,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&cno1,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&cno_blk,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&nmx_rop,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_rop,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_acf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_acf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_frf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_frf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_gcr_acf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_gcr_acf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_gcr_frf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_gcr_frf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

   MPI_Bcast(bs_sap,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmr_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncy_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nkv_sap,1,MPI_INT,0,MPI_COMM_WORLD);   
   
   if (nodfl==0)
   {
      MPI_Bcast(bs_dfl,4,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(&Ns,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rkappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&rinv,1,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(&rnmr,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rncy,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rvnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rvres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&vnkv,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&vnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&vres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&vdnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&vdres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }   
   
   start_ranlux(0,1234);
   geometry();
   alloc_ud();

   if (nodfl==0)
   {
      if ((cno0+cno1)==0)
         alloc_sd(7);
      else
         alloc_sd(8);
      
      if (Ns<(2*nkv_sap))
         alloc_s(2*nkv_sap+1);
      else
         alloc_s(Ns+2);

      alloc_u();
      alloc_sw();
      dfl=set_dfl_parms(bs_dfl,Ns,rinv,rnmr,rncy,rvnmx,vnkv,vnmx,vdnmx,
                        rkappa,rvres,vres,vdres,0,0);
   }
   else
   {
      if ((cno0+cno1)==0)
         alloc_sd(5);
      else
         alloc_sd(6);

      alloc_s(2*nkv_sap+1);
   }   

   lat=set_lat_parms(beta,kappa,csw);
   gcr=set_gcr_parms(bs_sap,nmr_sap,ncy_sap,nkv_sap);
   hmc=set_hmc_parms(bs_hmc,bc,N0,N1,N2[0],nodfl==0,cno0,cno1,cno_blk,
                     nmx_acf,nmx_gcr_acf,nmx_rop,nmx_frf,nmx_gcr_frf,
                     tau,
                     res_acf,res_gcr_acf,res_rop,res_frf,res_gcr_frf);

   if (my_rank==0)
   {
      printf("beta = %.6f\n",lat.beta);      
      printf("kappa = %.6f\n",lat.kappa);      
      printf("csw = %.6f\n",lat.csw);
      printf("bc = %d\n\n",hmc.bc);            

      printf("bs_hmc = %d %d %d %d\n",bs_hmc[0],bs_hmc[1],bs_hmc[2],bs_hmc[3]);
      printf("tau = %.2f\n",hmc.tau);
      printf("N0 = %d\n",hmc.N0);
      printf("N1 = %d\n",hmc.N1);
      printf("N2 = %d, %d, %d\n",hmc.N2,N2[1],N2[2]);      
      printf("cno0 = %d\n",hmc.cno0);
      printf("cno1 = %d\n",hmc.cno1);
      printf("cno_blk = %d\n",hmc.cno_blk);      
      printf("nmx_acf = %d\n",hmc.nmx_acf);            
      printf("res_acf = %.2e\n",hmc.res_acf);      
      printf("nmx_rop = %d\n",hmc.nmx_rop);            
      printf("res_rop = %.2e\n",hmc.res_rop);      
      printf("nmx_frf = %d\n",hmc.nmx_frf);            
      printf("res_frf = %.2e\n\n",hmc.res_frf);      

      printf("bs_sap = %d %d %d %d\n",bs_sap[0],bs_sap[1],bs_sap[2],bs_sap[3]);
      printf("nmr = %d\n",gcr.nmr);                  
      printf("ncy = %d\n",gcr.ncy);
      printf("nkv = %d\n",gcr.nkv);                        
      printf("nmx_gcr_acf = %d\n",hmc.nmx_gcr_acf);            
      printf("res_gcr_acf = %.2e\n",hmc.res_gcr_acf);
      printf("nmx_gcr_frf = %d\n",hmc.nmx_gcr_frf);            
      printf("res_gcr_frf = %.2e\n\n",hmc.res_gcr_frf);

      if (nodfl==0)
      {
         printf("bs_dfl = %d %d %d %d\n",
                bs_dfl[0],bs_dfl[1],bs_dfl[2],bs_dfl[3]);
         printf("Ns = %d\n",dfl.Ns);
         printf("rkappa = %.6f\n",dfl.rkappa);
         printf("rinv = %d\n",dfl.rinv);
         printf("rnmr = %d\n",dfl.rnmr);
         printf("rncy = %d\n",dfl.rncy);         
         printf("rvnmx = %d\n",dfl.rvnmx);
         printf("rvres = %.1e\n",dfl.rvres);
         printf("vnkv = %d\n",dfl.vnkv);         
         printf("vnmx = %d\n",dfl.vnmx);
         printf("vres = %.1e\n",dfl.vres);
         printf("vdnmx = %d\n",dfl.vdnmx);
         printf("vdres = %.1e\n\n",dfl.vdres);
      }

      printf("Configurations no %d -> %d in steps of %d\n\n",
             nc_first,nc_last,nc_step);
      fflush(flog);
   }

   alloc_hmc_action(&act0);
   alloc_hmc_action(&act1);

   error_root(((nc_last-nc_first)%nc_step)!=0,1,"main [check13.c]",
              "nc_last-nc_first is not a multiple of nc_step");
   check_dir_root(cnfg_dir);   

   nsize=name_size("%dx%dx%dx%db%.2fk%.5fc%.5fid%d",
                   NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);
   error_root(nsize>=NAME_SIZE,1,"main [check13.c]","Base name is too long");
   sprintf(nbase,"%dx%dx%dx%db%.2fk%.5fc%.5fid%d",
           NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);

   nsize=name_size("%s/%sn%d",cnfg_dir,nbase,nc_last);
   error_root(nsize>=NAME_SIZE,1,"main [check13.c]",
              "cnfg_dir name is too long");

   for (icnfg=nc_first;icnfg<=nc_last;icnfg+=nc_step)
   {
      if (my_rank==0)
      {
         printf("Configuration no %d:\n\n",icnfg);
         fflush(flog);
      }   

      sprintf(cnfg_file,"%s/%sn%d",cnfg_dir,nbase,icnfg);
      import_cnfg(cnfg_file);    

      hmc=set_hmc_parms(bs_hmc,bc,N0,N1,N2[0],nodfl==0,cno0,cno1,cno_blk,
                        nmx_acf,nmx_gcr_acf,nmx_rop,nmx_frf,nmx_gcr_frf,
                        tau,
                        res_acf,res_gcr_acf,res_rop,res_frf,res_gcr_frf);      
      start_md(act0,&stat_rop,&stat_dfl_mds,&ifail);
      sact0=sum_action(act0);
      save_mom();

      if (my_rank==0)
      {
         printf("start_md:\n");
         if (nodfl)
            printf("stat_rop = %d, ifail = %d\n",stat_rop,ifail);
         else
            printf("stat_rop = %d, stat_dfl_mds = %d, ifail = %d\n",
                   stat_rop,stat_dfl_mds,ifail);
         printf("action = % .4e\n\n",sact0);
         fflush(flog);
      }

      error_root((stat_rop<0)||(stat_dfl_mds<0)||(ifail!=0),1,
                 "main [check13.c]","start_md() failed");

      for (i=0;i<3;i++)
      {
         if (i>0)
            restart_md(N0,N1,N2[i]);
         run_md(act1,&stat_acf,stat_gcr_acf,&stat_frf,stat_gcr_frf,
                stat_dfl_upd,&ifail);
         sact1=sum_action(act1);
         dact[i]=diff_action(act0,act1);

         if (my_rank==0)
         {
            printf("run_md:\n");
            printf("N0,N1,N2 = %d,%d,%d\n",N0,N1,N2[i]);
            printf("stat_acf = %d\n",stat_acf);
            if (nodfl)
               printf("stat_gcr_acf = %d\n",stat_gcr_acf[0]);
            else
               printf("stat_gcr_acf = %d (little: %d, %d)\n",
                      stat_gcr_acf[0],stat_gcr_acf[1],stat_gcr_acf[2]);
            printf("stat_frf = %d\n",stat_frf);
            if (nodfl)
               printf("stat_gcr_frf = %d, ifail = %d\n",stat_gcr_frf[0],ifail);
            else
               printf("stat_gcr_frf = %d (little: %d, %d), "
                      "stat_dfl_upd = %d (little: %d), ifail = %d\n",
                      stat_gcr_frf[0],stat_gcr_frf[1],stat_gcr_frf[2],
                      stat_dfl_upd[0],stat_dfl_upd[1],ifail);
            printf("action = % .4e, dH = %.2e\n\n",sact1,dact[i]);      
            fflush(flog);
         }

         error_root((stat_acf<0)||(stat_gcr_acf[0]<0)||(stat_gcr_acf[1]<0)||
                    (stat_gcr_acf[2]<0)||(stat_frf<0)||(stat_gcr_frf[0]<0)||
                    (stat_gcr_frf[1]<0)||(stat_gcr_frf[2]<0)||
                    (stat_dfl_upd[1]<0)||(ifail!=0),1,"main [check13.c]",
                    "run_md() failed");
      }

      error_chk();
   
      if (my_rank==0)
      {
         eps0[0]=hmc.tau/(double)(N0*N1*N2[0]);
         printf("eps0=%.4f, dH = %.2e\n",eps0[0],dact[0]);

         for (i=1;i<3;i++)
         {
            eps0[i]=hmc.tau/(double)(N0*N1*N2[i]);
            printf("eps0=%.4f, dH = %.2e, |dH[i]/dH[i-1]| = %.2e (%.2e)\n",
                   eps0[i],dact[i],fabs(dact[i]/dact[i-1]),
                   (eps0[i]*eps0[i])/(eps0[i-1]*eps0[i-1]));
         }

         printf("\n");
         printf("(the theoretical scale factor is given in brackets)\n\n");
         fflush(flog);
      }
   }

   if (my_rank==0)
      fclose(flog);
      
   MPI_Finalize();    
   exit(0);
}
