
/*******************************************************************************
*
* File check5.c
*
* Copyright (C) 2005, 2008 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of hop_frc()
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "block.h"
#include "linalg.h"
#include "dirac.h"
#include "sw_term.h"
#include "hmc.h"
#include "global.h"

static const su3_dble ud0={{0.0}};


static void init_ud(void)
{
   su3_alg_dble X;
   su3_dble *u,*um;

   u=pud[VOLUME/2][0];
   um=u+4*VOLUME;

   for (;u<um;u++)
   {
      *u=ud0;
      (*u).c11.re=1.0;
      (*u).c22.re=1.0;
      (*u).c33.re=1.0;      
      random_alg(1,&X);
      ExpX(0.1,&X,u);      
   }

   set_flags(NEW_UD);
   renormalize_ud();
}


static void rot_ud(int n,double t)
{
   int nb,isw,*lk,*lkm;
   su3_dble *ub;
   su3_alg_dble *mom;
   block_t *b;
   hmc_t *hflds;

   b=blk_list(HMC_BLOCKS,&nb,&isw)+n;
   hflds=hmcflds()+n;
   
   ub=(*b).ud[0];
   mom=(*hflds).mom;
   lk=(*hflds).lks;
   lkm=lk+(*hflds).nlks;
   
   for (;lk<lkm;lk++)
   {
      ExpX(t,mom,ub+(*lk));
      mom+=1;
   }

   set_blk_flags(HMC_BLOCKS,NEW_ACTIVE_UD,n);
}


static double dSdt(int n)
{
   double sm;
   su3_alg_dble *fr,*frm,*mom;
   hmc_t *hflds;

   hflds=hmcflds()+n;

   fr=(*hflds).frc;
   frm=fr+(*hflds).nlks;
   mom=(*hflds).mom;   
   sm=0.0;
   
   for (;fr<frm;fr++)
   {
      sm+=12.0*((*mom).c1*(*fr).c1+(*mom).c2*(*fr).c2);
      sm-= 6.0*((*mom).c1*(*fr).c2+(*mom).c2*(*fr).c1);
      sm+= 4.0*((*mom).c3*(*fr).c3+(*mom).c4*(*fr).c4+(*mom).c5*(*fr).c5+
                (*mom).c6*(*fr).c6+(*mom).c7*(*fr).c7+(*mom).c8*(*fr).c8);
      mom+=1;
   }

   return sm;
}


static double hop_act(int n)
{
   int nb,isw;
   double act;
   block_t *b;

   b=blk_list(HMC_BLOCKS,&nb,&isw)+n;

   Qoe_blk_dble(b,0,3);
   act=spinor_prod_re_dble((*b).volh,0,
                           (*b).sd[3]+(*b).volh,(*b).sd[1]+(*b).volh);

   Qoe_blk_dble(b,1,3);
   act+=spinor_prod_re_dble((*b).volh,0,
                            (*b).sd[0]+(*b).volh,(*b).sd[3]+(*b).volh);

   return 2.0*act;
}


int main(int argc,char *argv[])
{
   int my_rank;
   int bs[4],n,nb,isw;
   double eps,dsdt,act0,act1,dact;
   double dev_frc,dev_frc_all,dev_frc_max;
   double sig_loss,sig_loss_all,sig_loss_max;
   block_t *b;
   hmc_t *hflds;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check5.log","w",stdout);
      fin=freopen("check5.in","r",stdin);
      
      printf("\n");
      printf("Check of hop_frc()\n");
      printf("------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bs","%d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);      
      fclose(fin);

      printf("bs = %d %d %d %d\n",bs[0],bs[1],bs[2],bs[3]);      
      printf("Random gauge field\n\n");
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);

   set_lat_parms(0.0,0.1,1.234);
   start_ranlux(0,1234);
   geometry();
   alloc_ud();
   alloc_bgr(HMC_BLOCKS,bs);

   init_ud();
   assign_ud2udbgr(HMC_BLOCKS);

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();   

   dev_frc_max=0.0;
   sig_loss_max=0.0;
   eps=3.0e-4;
   
   for (n=0;n<nb;n++)
   {
      random_alg((*hflds).nlks,(*hflds).mom);
      random_sd((*b).vol,(*b).sd[0],1.0);
      random_sd((*b).vol,(*b).sd[1],1.0);
      
      set_alg2zero((*hflds).nlks,(*hflds).frc);
      hop_frc(n,0,1);
      dsdt=dSdt(n);

      rot_ud(n,eps);
      act0=2.0*hop_act(n)/3.0;
      rot_ud(n,-eps);

      rot_ud(n,-eps);
      act1=2.0*hop_act(n)/3.0;
      rot_ud(n,eps);
      
      rot_ud(n,2.0*eps);
      act0-=hop_act(n)/12.0;
      rot_ud(n,-2.0*eps);

      rot_ud(n,-2.0*eps);
      act1-=hop_act(n)/12.0;
      rot_ud(n,2.0*eps);

      dact=(act0-act1)/eps;
      dev_frc=fabs(1.0-dsdt/dact);
      if (dev_frc>dev_frc_max)
         dev_frc_max=dev_frc;

      sig_loss=-log10(fabs(1.0-act0/act1));
      if (sig_loss>sig_loss_max)
         sig_loss_max=sig_loss;      

      b+=1;
      hflds+=1;
   }

   MPI_Reduce(&dev_frc_max,&dev_frc_all,1,
              MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
   MPI_Reduce(&sig_loss_max,&sig_loss_all,1,
              MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
   error_chk();

   if (my_rank==0)
   {
      printf("Calculation of the force:\n");
      printf("Relative deviation of dS/dt = %.2e ",dev_frc_all);
      printf("[significance loss = %d digits]\n\n",(int)(sig_loss_all)); 
      fclose(flog);
   }
   
   MPI_Finalize();    
   exit(0);
}
