
/*******************************************************************************
*
* File check4.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Consistency checks on the programs in the module valg_dble
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "linalg.h"
#include "dfl.h"
#include "global.h"

static complex_dble v[25];
static complex_dble *ppk[5];


static complex_dble sp(int vol,complex_dble *pk,complex_dble *pl)
{
   int ix;
   double x,y;
   complex_dble z;

   x=0.0;
   y=0.0;

   for (ix=0;ix<vol;ix++)
   {
      x+=(double)((*pk).re*(*pl).re+(*pk).im*(*pl).im);
      y+=(double)((*pk).re*(*pl).im-(*pk).im*(*pl).re);
      pk+=1;
      pl+=1;
   }
   
   z.re=x;
   z.im=y;
   
   return z;
}


int main(int argc,char *argv[])
{
   int my_rank,i,j,vol,off;
   int bs[4],nm,n,nv,nvec;
   int icom,ieo;
   double r,zsq;
   double d,dmax;
   complex_dble z,w;
   complex_dble **wv,*pk,*pl;
   FILE *flog=NULL,*fin=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check4.log","w",stdout);
      fin=freopen("check4.in","r",stdin);
      
      printf("\n");
      printf("Consistency of the programs in the module valg_dble\n");
      printf("---------------------------------------------------\n\n");   

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bs","%d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);
      read_line("nm","%d",&nm);
      fclose(fin);

      printf("bs = %d %d %d %d\n",bs[0],bs[1],bs[2],bs[3]);
      printf("nm = %d\n\n",nm);
      fflush(flog);
   }      

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nm,1,MPI_INT,0,MPI_COMM_WORLD);   

   start_ranlux(0,12345);
   geometry();
   set_dfl_parms(bs,nm,4,1,1,1,1,1,1,0.1,1.0e-3,1.0e-6,1.0e-12,0,10);   
   wv=reserve_wvd(10);
   vdflds_base(&n,&nv,&nvec);
   
   for (icom=0;icom<2;icom++)
   {
      if ((icom==0)||(NPROC>1))
      {
         if (my_rank==0)
         {
            if (icom==1)
            {
               printf("Checks with global summation\n");
               printf("============================\n\n");
            }
            else
            {
               printf("Checks without global summation\n");
               printf("===============================\n\n");
            }
         }
      
         for (ieo=0;ieo<3;ieo++)
         {
            if (my_rank==0)
            {
               if (ieo==0)
                  printf("First case: full lattice\n\n");
               else if (ieo==1)
                  printf("Second case: even points\n\n");
               else
                  printf("Third case: odd points\n\n");
            }

            vol=nv/2;
            off=0;

            if (ieo==0)
               vol=nv;
            if (ieo==2)
               off=nv/2;

            for (i=0;i<10;i++)
               random_vd(vol,wv[i]+off,1.0f);

            dmax=0.0;
   
            for (i=0;i<10;i++)
            {
               pk=wv[i]+off;
               pl=wv[9-i]+off;

               if (icom==1)
               {
                  z=sp(vol,pk,pl);
                  MPI_Reduce(&z.re,&w.re,2,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
                  MPI_Bcast(&w.re,2,MPI_DOUBLE,0,MPI_COMM_WORLD); 
               }
               else
                  w=sp(vol,pk,pl);

               z=vprod_dble(vol,icom,pk,pl);
               r=vnorm_square_dble(vol,icom,pk)*vnorm_square_dble(vol,icom,pl);
               d=(z.re-w.re)*(z.re-w.re)+(z.im-w.im)*(z.im-w.im);
               d=sqrt(d/r);
               if (d>dmax)
                  dmax=d;

               z=vprod_dble(vol,icom,pk,pk);
               r=vnorm_square_dble(vol,icom,pk);
      
               d=fabs(z.im/r);
               if (d>dmax)
                  dmax=d;

               d=fabs(z.re/r-1.0);
               if (d>dmax)
                  dmax=d;
            }

            if (my_rank==0)
               printf("Check of vprod_dble and vnorm_square_dble: %.2e\n\n",
                      dmax);
   
            dmax=0.0;
            z.re= 0.345;
            z.im=-0.876;
            zsq=z.re*z.re+z.im*z.im;
   
            for (i=0;i<9;i++)
            {
               pk=wv[i]+off;
               pl=wv[i+1]+off;      
      
               w=vprod_dble(vol,icom,pk,pl);
               r=vnorm_square_dble(vol,icom,pk)+
                  zsq*vnorm_square_dble(vol,icom,pl)
                  +2.0f*(z.re*w.re-z.im*w.im);
               mulc_vadd_dble(vol,pk,pl,z);

               d=fabs(r/vnorm_square_dble(vol,icom,pk)-1.0);
               if (d>dmax)
                  dmax=d;
            }

            if (my_rank==0)
            {   
               printf("Consistency of vprod_dble, vnorm_square_dble\n");
               printf("and mulc_vadd_dble: %.2e\n\n",dmax);
            }
   
            for (i=0;i<10;i++)
               random_vd(vol,wv[i]+off,1.0f);

            dmax=0.0;
   
            for (i=0;i<10;i++)
            {
               pk=wv[i]+off;
      
               if (i>0)
               {
                  pl=wv[i-1]+off;         
                  vproject_dble(vol,icom,pk,pl);
                  z=vprod_dble(vol,icom,pk,pl);

                  d=(fabs(z.re)+fabs(z.im))/
                     sqrt(vnorm_square_dble(vol,icom,pk));

                  if (d>dmax)
                     dmax=d;
               }

               vnormalize_dble(vol,icom,pk);
               r=vnorm_square_dble(vol,icom,pk);
               
               d=fabs(r-1.0);
               if (d>dmax)
                  dmax=d;
            }

            if (my_rank==0)
            {   
               printf("Consistency of vprod_dble, vnorm_square_dble,\n");
               printf("vnormalize_dble and vproject_dble: %.2e\n\n",dmax);
            }
   
            for (i=0;i<5;i++)
            {
               pk=wv[i]+off;
               pl=wv[i+5]+off;
      
               random_vd(vol,wv[i]+off,1.0f);
               assign_vd2vd(vol,pk,pl);

               for (j=0;j<5;j++)
               {
                  v[5*i+j].re=0.1234*(double)(i^2)-0.8976*(double)(j);
                  v[5*i+j].im=0.2231*(double)(i)+0.9922*(double)(j^2);
               }

               ppk[i]=pl;
            }

            vrotate_dble(vol,5,ppk,v);
            dmax=0.0;
   
            for (i=5;i<10;i++)
            {
               pk=wv[i]+off;

               for (j=0;j<5;j++)
               {
                  z.re=-v[5*j+(i-5)].re;
                  z.im=-v[5*j+(i-5)].im;

                  pl=wv[j]+off;         
                  mulc_vadd_dble(vol,pk,pl,z);
               }

               r=vnorm_square_dble(vol,icom,pk);

               d=fabs(r);
               if (d>dmax)
                  dmax=d;
            }

            dmax/=vnorm_square_dble(vol,icom,wv[0]+off);
            dmax=sqrt(dmax);

            if (my_rank==0)
            {
               printf("Consistency of mulc_vadd_dble\n");
               printf("and vrotate_dble: %.2e\n\n",dmax);
            }
         }
      }
   }

   error_chk();
   
   if (my_rank==0)
      fclose(flog);
   MPI_Finalize();   
   exit(0);
}
