
/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2005, 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Writing and reading configurations
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "misc.h"
#include "start.h"
#include "update.h"
#include "global.h"

static float rs1[100],rs2[100];
static double rd1[100],rd2[100];
static su3_dble u[4*VOLUME];


int main(int argc,char *argv[])
{
   int my_rank,nsize,i;
   float beta;
   double *ru,*rv,d,dmax;
   su3_dble *v;
   char loc_dir[NAME_SIZE],name[NAME_SIZE];
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check2.log","w",stdout);
      fin=freopen("check2.in","r",stdin);
      
      printf("\n");
      printf("Writing and reading configurations\n");
      printf("----------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("loc_dir","%s\n",loc_dir);
      fclose(fin);
      
      fflush(flog);
   }

   MPI_Bcast(loc_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);   

   beta=5.7f;
   start_ranlux(0,123456);
   geometry();
   alloc_u();
   alloc_ud();
   random_u();

   update(beta,4,0,8);
   assign_u2ud();

   check_dir(loc_dir);   
   nsize=name_size("%s/testcnfg_%d",loc_dir,NPROC);
   error_root(nsize>=NAME_SIZE,1,"main [check2.c]","loc_dir name is too long");
   sprintf(name,"%s/testcnfg_%d",loc_dir,my_rank);   
   write_cnfg(name);

   ranlxs(rs1,100);
   ranlxd(rd1,100);
   v=pud[VOLUME/2][0];

   for (i=0;i<(4*VOLUME);i++)
      u[i]=v[i];

   update(beta,4,0,8);
   assign_u2ud();

   read_cnfg(name);
   remove(name);

   ranlxs(rs2,100);
   ranlxd(rd2,100);

   dmax=0.0;

   for (i=0;i<100;i++)
   {
      d=fabs((double)(rs1[i]-rs2[i]));
      if (d>dmax)
         dmax=d;

      d=fabs(rd1[i]-rd2[i]);
      if (d>dmax)
         dmax=d;
   }

   ru=(double*)(u);
   rv=(double*)(v);

   for (i=0;i<(72*VOLUME);i++)
   {
      d=fabs((*ru)-(*rv));
      if (d>dmax)
         dmax=d;
      ru+=1;
      rv+=1;
   }

   error_chk();
   MPI_Reduce(&dmax,&d,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);

   if (my_rank==0)
   {
      printf("Wrote gauge configuration to the files %s/testcnfg_*\n"
             "on the local disks. ",loc_dir);
      printf("Then read the configuration from there and\n"
             "removed the files\n\n");
      
      printf("Maximal deviation = %.1e ",d);
      printf("(should be exactly equal to 0.0)\n\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
