
/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of spin2mat
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "su3.h"
#include "random.h"
#include "start.h"
#include "sw_term.h"
#include "hmc.h"

#if (defined SSE)
static su3_dble u,v __attribute__ ((aligned (16)));
static spinor_dble rx,ry,sx,sy,sw __attribute__ ((aligned (16)));
#else
static su3_dble u,v;
static spinor_dble rx,ry,sx,sy,sw;
#endif

static su3_dble ud0={{0.0}};

#define _re(z,w) ((z).re*(w).re+(z).im*(w).im)
#define _im(z,w) ((z).im*(w).re-(z).re*(w).im)


static su3_vector_dble mul_cplx(complex_dble z,su3_vector_dble s)
{
   su3_vector_dble r;

   r.c1.re=z.re*s.c1.re-z.im*s.c1.im;
   r.c1.im=z.im*s.c1.re+z.re*s.c1.im;
   r.c2.re=z.re*s.c2.re-z.im*s.c2.im;
   r.c2.im=z.im*s.c2.re+z.re*s.c2.im;
   r.c3.re=z.re*s.c3.re-z.im*s.c3.im;
   r.c3.im=z.im*s.c3.re+z.re*s.c3.im;

   return r;
}


static spinor_dble mul_gamma(int mu,spinor_dble s)
{
   spinor_dble r;
   complex_dble i,m_i,m_1;

   i.re=0.0;
   i.im=1.0;

   m_i.re=0.0;
   m_i.im=-1.0;

   m_1.re=-1.0;
   m_1.im=0.0;

   if (mu==0)
   {
      r.c1=mul_cplx(m_1,s.c3);
      r.c2=mul_cplx(m_1,s.c4);
      r.c3=mul_cplx(m_1,s.c1);
      r.c4=mul_cplx(m_1,s.c2);
   }
   else if (mu==1)
   {
      r.c1=mul_cplx(m_i,s.c4);
      r.c2=mul_cplx(m_i,s.c3);
      r.c3=mul_cplx(i,s.c2);
      r.c4=mul_cplx(i,s.c1);
   }
   else if (mu==2)
   {
      r.c1=mul_cplx(m_1,s.c4);
      r.c2=s.c3;
      r.c3=s.c2;
      r.c4=mul_cplx(m_1,s.c1);
   }
   else if (mu==3)
   {
      r.c1=mul_cplx(m_i,s.c3);
      r.c2=mul_cplx(i,s.c4);
      r.c3=mul_cplx(i,s.c1);
      r.c4=mul_cplx(m_i,s.c2);
   }
   else
   {
      r.c1=s.c1;
      r.c2=s.c2;
      r.c3=mul_cplx(m_1,s.c3);
      r.c4=mul_cplx(m_1,s.c4);
   }

   return r;
}


static void add_tensor(su3_vector_dble *r,su3_vector_dble *s,su3_dble *p)
{
   (*p).c11.re+=_re((*r).c1,(*s).c1);
   (*p).c11.im+=_im((*r).c1,(*s).c1);   
   (*p).c12.re+=_re((*r).c1,(*s).c2);
   (*p).c12.im+=_im((*r).c1,(*s).c2);   
   (*p).c13.re+=_re((*r).c1,(*s).c3);
   (*p).c13.im+=_im((*r).c1,(*s).c3);   

   (*p).c21.re+=_re((*r).c2,(*s).c1);
   (*p).c21.im+=_im((*r).c2,(*s).c1);   
   (*p).c22.re+=_re((*r).c2,(*s).c2);
   (*p).c22.im+=_im((*r).c2,(*s).c2);   
   (*p).c23.re+=_re((*r).c2,(*s).c3);
   (*p).c23.im+=_im((*r).c2,(*s).c3);   

   (*p).c31.re+=_re((*r).c3,(*s).c1);
   (*p).c31.im+=_im((*r).c3,(*s).c1);   
   (*p).c32.re+=_re((*r).c3,(*s).c2);
   (*p).c32.im+=_im((*r).c3,(*s).c2);   
   (*p).c33.re+=_re((*r).c3,(*s).c3);
   (*p).c33.im+=_im((*r).c3,(*s).c3);   
}


static double max_dev(su3_dble *u,su3_dble *v)
{
   complex_dble *r,*s,*rm;
   double nrm,d,dmax;

   r=(complex_dble*)(&(*u).c11);
   s=(complex_dble*)(&(*v).c11);
   rm=r+9;

   nrm=0.0;
   dmax=0.0;
   
   for (;r<rm;r++)
   {
      nrm+=(*r).re*(*r).re+(*r).im*(*r).im;

      d=((*r).re-(*s).re)*((*r).re-(*s).re)+
        ((*r).im-(*s).im)*((*r).im-(*s).im);
      if (d>dmax)
         dmax=d;

      s+=1;
   }

   return sqrt(dmax/nrm);
}


int main(void)
{
   int mu;
   
   printf("\n");
   printf("Check of spin2mat\n");
   printf("-----------------\n\n");

   rlxd_init(1,567);

   gauss_dble(&rx.c1.c1.re,24);
   gauss_dble(&ry.c1.c1.re,24);
   gauss_dble(&sx.c1.c1.re,24);
   gauss_dble(&sy.c1.c1.re,24);   
   
   for (mu=0;mu<4;mu++)
   {
      spin2mat[mu](&rx,&ry,&sx,&sy,&u);

      v=ud0;

      sw=mul_gamma(mu,ry);
      _vector_sub(sw.c1,ry.c1,sw.c1); 
      _vector_sub(sw.c2,ry.c2,sw.c2);       
      _vector_sub(sw.c3,ry.c3,sw.c3);
      _vector_sub(sw.c4,ry.c4,sw.c4);             
      sw=mul_gamma(5,sw);

      add_tensor(&sw.c1,&sx.c1,&v);
      add_tensor(&sw.c2,&sx.c2,&v);
      add_tensor(&sw.c3,&sx.c3,&v);
      add_tensor(&sw.c4,&sx.c4,&v);

      sw=mul_gamma(mu,sy);
      _vector_sub(sw.c1,sy.c1,sw.c1); 
      _vector_sub(sw.c2,sy.c2,sw.c2);       
      _vector_sub(sw.c3,sy.c3,sw.c3);
      _vector_sub(sw.c4,sy.c4,sw.c4);             
      sw=mul_gamma(5,sw);

      add_tensor(&sw.c1,&rx.c1,&v);
      add_tensor(&sw.c2,&rx.c2,&v);
      add_tensor(&sw.c3,&rx.c3,&v);
      add_tensor(&sw.c4,&rx.c4,&v);

      printf("mu = %d: %.2e\n",mu,max_dev(&u,&v));
   }

   printf("\n");
   exit(0);
}
