
/*******************************************************************************
*
* File cheby2.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Chebyshev approximation of sin(pi*x/2)=x*sum_n c_n*T_n(x),  -1.0<=x<=1.0
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "extras.h"


static double f(double x)
{
   double z,y;

   z=2.0*atan(1.0)*x;
   
   if (fabs(z)<=1.0e-3)
   {
      y=1.0/120.0;
      y=z*z*y-1.0/6.0;
      y=z*z*y+1.0;
      y=2.0*atan(1.0)*y;
   }
   else
      y=sin(z)/x;

   return(y);
}


int main(void)
{
   int nmax=64,n,i;
   double a,b,eps,c[64];

   printf("\n");
   printf("Chebyshev approximation of sin(pi*x/2)/x\n");
   printf("----------------------------------------\n");
   printf("\n");

   a=-1.0;
   b=1.0;
   eps=1.0e-8;

   printf("Range %.3f<=x<=%.3f, absolute accuracy %.1e\n\n",a,b,eps);
   
   n=cheby_fit(a,b,f,nmax,eps,c);

   for (i=0;i<=n;i++)
      printf("   c[%d]=% 10.9ff;\n",i,c[i]);

   printf("\n");   
   exit(0);
}
