
/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Decomposition of floating point numbers into 2^p*1.f using the SSE unit
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "random.h"

#if (defined SSE)

#include "sse.h"
#define N 4000000

static sse_int mask1,mask2,mask3,p;
static sse_float x,y,f;
static float r[4*N] __attribute__ ((aligned (16)));


int main(void)
{
   int i,itest;
   
   printf("\n");
   printf("Decomposition of floating point numbers in 2^p*1.f\n");
   printf("--------------------------------------------------\n");
   printf("\n");

   mask1.c1=0x7f800000;
   mask1.c2=0x7f800000;
   mask1.c3=0x7f800000;
   mask1.c4=0x7f800000;

   mask2.c1=0x7fffff;
   mask2.c2=0x7fffff;
   mask2.c3=0x7fffff;
   mask2.c4=0x7fffff;

   mask3.c1=0x3f800000;
   mask3.c2=0x3f800000;
   mask3.c3=0x3f800000;
   mask3.c4=0x3f800000;   
   
   ranlxs(r,4*N);
   itest=0;

   for (i=0;i<4*N;i+=4)
   {
      x.c1=(float)(tan(1.57*(double)(1.0f-r[i])));
      x.c2=(float)(tan(1.57*(double)(1.0f-r[i+1])));
      x.c3=(float)(tan(1.57*(double)(1.0f-r[i+2])));
      x.c4=(float)(tan(1.57*(double)(1.0f-r[i+3])));
      
      __asm__ __volatile__ ("movaps %2, %%xmm0 \n\t"
                            "movaps %2, %%xmm1 \n\t"
                            "andps %3, %%xmm0 \n\t"                            
                            "andps %4, %%xmm1 \n\t"
                            "orps %5, %%xmm1 \n\t"
                            "movaps %%xmm0, %0 \n\t"
                            "movaps %%xmm1, %1"
                            :
                            "=m" (p),
                            "=m" (f)
                            :
                            "m" (x),
                            "m" (mask1),
                            "m" (mask2),
                            "m" (mask3));

      p.c1=(p.c1>>23)-127;
      p.c2=(p.c2>>23)-127;
      p.c3=(p.c3>>23)-127;   
      p.c4=(p.c4>>23)-127;

      y.c1=(float)(ldexp(f.c1,p.c1));
      y.c2=(float)(ldexp(f.c2,p.c2));
      y.c3=(float)(ldexp(f.c3,p.c3));
      y.c4=(float)(ldexp(f.c4,p.c4));      

      itest+=(x.c1!=y.c1);
      itest+=(x.c2!=y.c2);
      itest+=(x.c3!=y.c3);
      itest+=(x.c4!=y.c4);      
   }
   
   printf("Number of tested numbers: %d\n",4*N);
   printf("Number of wrong decompositions: %d\n\n",itest);
   exit(0);
}

#else

int main(void)
{
   printf("\n");
   printf("Decomposition of floating point numbers in 2^p*1.f\n");
   printf("--------------------------------------------------\n\n");

   printf("SSE must be enabled for this test\n\n");
   exit(0);
}

#endif
