
/*******************************************************************************
*
* File check3.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Test of the Chebyshev approximation of sin(pi*x/2)/x 
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "random.h"

#define N 1000000

static double f(double x)
{
   double z,y;

   z=2.0*atan(1.0)*x;
   
   if (fabs(z)<=1.0e-3)
   {
      y=1.0/120.0;
      y=z*z*y-1.0/6.0;
      y=z*z*y+1.0;
      y=2.0*atan(1.0)*y;
   }
   else
      y=sin(z)/x;

   return(y);
}


int main(void)
{
   int i;
   float c[5],r[N],u,v,w,x,y,z;
   double d,dmax;

   printf("\n");
   printf("Check of the Chebyshev approximation of sin(pi*x/2)/x\n");
   printf("-----------------------------------------------------\n");
   printf("\n");

   c[0]= 1.276278962f;
   c[1]=-0.285261569f;
   c[2]= 0.009118016f;
   c[3]=-0.000136588f;
   c[4]= 0.000001185f;
   
   ranlxs(r,N);

   dmax=0.0;

   for (i=0;i<N;i++)
   {
      x=-1.0f+2.0f*r[i];

      z=4.0f*x*x-2.0f;
      u=c[4];
      v=c[3];
      w=z*u+v;
      v=u-c[2];
      u=z*w-v;
      v=w-c[1];
      w=z*u-v;
      v=u-c[0];
      y=0.5f*z*w-v;      

      d=f((double)(x))/(double)(y)-1.0;
      d=fabs(d);
      if (d>dmax)
         dmax=d;
   }
   
   printf("Range -1.0<=x<=1.0, %d uniformly distributed values\n",N);
   printf("Maximal relative error of the fit: %4.2e\n\n",dmax);   
   exit(0);
}
