
/*******************************************************************************
*
* File check9.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Test of the endianness and byte swapping programs
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "su3.h"
#include "random.h"
#include "start.h"
#include "misc.h"

#define N 100000

static int istd[N],istds[N];
static double dstd[N],dstds[N];


int main(void)
{
   int ie,k,it;
   stdint_t i[2];
   double d[2];
   char *ci[2],*cd[2];
   
   printf("\n");
   printf("Test of the endianness and byte swapping programs\n");
   printf("-------------------------------------------------\n\n");

   printf("sizeof(stdint_t) = %d\n",sizeof(stdint_t));
   printf("sizeof(double) = %d\n",sizeof(double));

   ie=endianness();
   if (ie==LITTLE_ENDIAN)
      printf("The machine is little endian\n\n");
   else if (ie==BIG_ENDIAN)
      printf("The machine is big endian\n\n");
   else
      printf("The machine has unknown endianness\n\n");

   ci[0]=(char*)(i);
   ci[1]=(char*)(i+1);

   ci[0][0]='A';
   ci[0][1]='B';
   ci[0][2]='C';
   ci[0][3]='D';   
   
   ci[1][0]='1';
   ci[1][1]='2';
   ci[1][2]='3';
   ci[1][3]='4';   

   printf("Byte swapping integers:\n");
   printf("%.4s, %.4s  ->  ",ci[0],ci[1]);
   bswap_int(2,i);
   printf("%.4s, %.4s\n\n",ci[0],ci[1]);

   cd[0]=(char*)(d);
   cd[1]=(char*)(d+1);

   cd[0][0]='A';
   cd[0][1]='B';
   cd[0][2]='C';
   cd[0][3]='D';
   cd[0][4]='E';
   cd[0][5]='F';
   cd[0][6]='G';
   cd[0][7]='H';    

   cd[1][0]='1';
   cd[1][1]='2';
   cd[1][2]='3';
   cd[1][3]='4';
   cd[1][4]='5';
   cd[1][5]='6';
   cd[1][6]='7';
   cd[1][7]='8';    

   printf("Byte swapping double precision numbers:\n");
   printf("%.8s, %.8s  ->  ",cd[0],cd[1]);
   bswap_double(2,d);
   printf("%.8s, %.8s\n\n",cd[0],cd[1]);

   gauss_dble(dstd,N);

   for (k=0;k<N;k++)
   {
      dstds[k]=dstd[k];
      istd[k]=(int)(dstd[k]);
      istds[k]=istd[k];
   }

   bswap_int(N,istd);
   bswap_int(N,istd);

   it=0;

   for (k=0;k<N;k++)
      if ((istds[k]-istd[k])!=0)
         it+=1;

   printf("2x(byte swap) %d integer random numbers:\n",N);
   printf("Number of failures = %d\n\n",it);
   
   bswap_double(N,dstd);
   bswap_double(N,dstd);

   it=0;

   for (k=0;k<N;k++)
      if ((dstds[k]-dstd[k])!=0.0)
         it+=1;

   printf("2x(byte swap) %d double-precision random numbers:\n",N);
   printf("Number of failures = %d\n\n",it);
   
   exit(0);
}
