
/*******************************************************************************
*
* File time5.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of apply_sw_dble
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "random.h"
#include "sw_term.h"
#include "start.h"

#define N 50000

int main(void)
{
   int i,n,count;
   double t1,t2,dt;
   spinor_dble *sb,*s;
   pauli_dble *mb,*m;

   printf("\n");
   printf("Timing of apply_sw_dble\n");
   printf("-----------------------\n\n");

   sb=amalloc(N*sizeof(spinor_dble),7);
   mb=amalloc(2*N*sizeof(pauli_dble),7);
   error((sb==NULL)||(mb==NULL),1,"main [time5.c]",
         "Unable to allocate auxiliary arrays");

   rlxd_init(1,23456);

   for (s=sb;s<(sb+N);s++)
      ranlxd((double*)(s),24);

   for (m=mb;m<(mb+2*N);m++)
   {
      ranlxd((*m).u,36);

      for (i=0;i<36;i++)
      {
         if (i<6)
            (*m).u[i]=1.0+1.0e-4*(*m).u[i];
         else
            (*m).u[i]*=1.0e-4;
      }
   }
   
   n=5000000/N;
   if (n<1)
      n=1;
   
   t1=(double)clock();
   for (count=0;count<n;count++)
      apply_sw_dble(N,mb,sb,sb);
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(2*N*n);

#if (defined SSE3)
   printf("Using SSE3 instructions:\n");
#elif (defined SSE2)
   printf("Using SSE2 instructions:\n");
#elif (defined SSE)
   printf("Using SSE instructions:\n");
#endif

   printf("The time per multiplication is %4.3f micro sec",dt);
   printf(" (%d Mflops [%d bit arithmetic])\n\n",
          (int)(276.0/dt),(int)(sizeof(spinor_dble)/3));

   exit(0);
}
