
/*******************************************************************************
*
* File check1.c
*
* Copyright (C) 2005, 2008 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the block inversion program blk_mres
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "linalg.h"
#include "block.h"
#include "sw_term.h"
#include "dirac.h"
#include "sap_gcr.h"
#include "global.h"


static void ref_mres(block_t *b,int nmr)
{
   int n,nsp;
   float rn;
   complex z;
   spinor *psx,*psr,*psar;

   nsp=(*b).volh;
   psx=(*b).s[0];
   psr=(*b).s[1];
   psar=(*b).s[2];

   set_s2zero(nsp,psx);

   for (n=0;n<nmr;n++)
   {
      Qhat_blk(b,1,2);
      mulg5(nsp,psar);
      rn=norm_square(nsp,0,psar);

      if (rn<(2.0f*FLT_MIN))
         return;

      z=spinor_prod(nsp,0,psar,psr);

      rn=1.0f/rn;
      z.re*=rn;
      z.im*=rn;
      mulc_spinor_add(nsp,psx,psr,z);

      z.re=-z.re;
      z.im=-z.im;
      mulc_spinor_add(nsp,psr,psar,z);
   }
}


int main(int argc,char *argv[])
{
   int my_rank,n;
   int bs[4],nmr,nb,isw,vol,volh;
   float r0,r1,r2,r0m,r1m,r2m;
   complex z;
   spinor *ps0,*ps1,*ps2;
   block_t *b0,*b,*bm;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check1.log","w",stdout);
      fin=freopen("check1.in","r",stdin);

      printf("\n");
      printf("Check of the block inversion program blk_mres\n");
      printf("---------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bs","%d %d %d %d\n",&bs[0],&bs[1],&bs[2],&bs[3]);
      read_line("nmr","%d",&nmr);
      fclose(fin);

      printf("bs = %d %d %d %d\n",bs[0],bs[1],bs[2],bs[3]);
      printf("nmr = %d\n\n",nmr);
      fflush(flog);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmr,1,MPI_INT,0,MPI_COMM_WORLD);

   start_ranlux(0,12345);
   geometry();
   alloc_ud();
   alloc_swd();
   alloc_s(3);
   alloc_bgr(GCR_BLOCKS,bs);

   set_lat_parms(0.0,0.1,0.2);
   z.re=-1.0f;
   z.im= 0.0f;
   
   random_ud();
   sw_term();
   assign_ud2ubgr(GCR_BLOCKS);
   n=assign_swd2swbgr(GCR_BLOCKS,ODD_PTS);
   error(n!=0,1,"main [check1.c]","Inversion of the SW term was not safe");

   b0=blk_list(GCR_BLOCKS,&nb,&isw);
   bm=b0+nb;
   r0m=0.0f;
   r1m=0.0f;
   r2m=0.0f;

   for (b=b0;b<bm;b++)
   {
      vol=(*b).vol;
      volh=(*b).volh;
      ps0=(*b).s[0];
      ps1=(*b).s[1];
      ps2=(*b).s[2];

      random_s(volh,ps1,1.0f);
      normalize(volh,0,ps1);
      assign_sb2s(b,EVEN_PTS,1,1);
      blk_mres(b,nmr);
      assign_sb2s(b,EVEN_PTS,0,0);
      assign_sb2s(b,EVEN_PTS,1,2);
      
      r0=norm_square(volh,0,ps1);
      if (r0>r0m)
         r0m=r0;

      Qhat_blk(b,0,2);
      mulg5(volh,ps2);
      assign_s2sb(b,EVEN_PTS,1,1);      
      mulc_spinor_add(volh,ps1,ps2,z);
      assign_s2sb(b,EVEN_PTS,2,2);
      mulc_spinor_add(volh,ps1,ps2,z);
      
      r1=norm_square(volh,0,ps1);
      if (r1>r1m)
         r1m=r1;

      assign_s2sb(b,EVEN_PTS,1,1);      
      ref_mres(b,nmr);
      assign_s2sb(b,EVEN_PTS,0,1);
      mulc_spinor_add(volh,ps0,ps1,z);

      r2=norm_square(volh,0,ps0);
      if (r2>r2m)
         r2m=r2;
   }

   error_chk();
   
   MPI_Reduce(&r0m,&r0,1,MPI_FLOAT, MPI_MAX,0,MPI_COMM_WORLD);
   MPI_Reduce(&r1m,&r1,1,MPI_FLOAT, MPI_MAX,0,MPI_COMM_WORLD);
   MPI_Reduce(&r2m,&r2,1,MPI_FLOAT, MPI_MAX,0,MPI_COMM_WORLD);   
   
   if (my_rank==0)
   {
      r0=(float)(sqrt((double)(r0)));
      r1=(float)(sqrt((double)(r1)));
      r2=(float)(sqrt((double)(r2)));

      printf("After %d MR iterations:\n",nmr);
      printf("Normalized residue = %.2e\n",r0);
      printf("Residue check      = %.2e\n",r1);
      printf("Algorithm check    = %.2e\n\n",r2);
      printf("(values maximized over all blocks and processes)\n\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
