
/*******************************************************************************
*
* File time2.c
*
* Copyright (C) 2005, 2008 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of the Schwarz preconditioner
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "misc.h"
#include "start.h"
#include "update.h"
#include "linalg.h"
#include "block.h"
#include "sw_term.h"
#include "dirac.h"
#include "sap_gcr.h"
#include "global.h"

static int my_rank,nmr,ncy;


static void Opz(spinor *pr,spinor *paz)
{
   spinor *rsm;

   rsm=pr+VOLUME;

   for (;pr<rsm;pr++)
   {
      _vector_sub((*paz).c1,(*pr).c1,(*paz).c1);
      _vector_sub((*paz).c2,(*pr).c2,(*paz).c2);
      _vector_sub((*paz).c3,(*pr).c3,(*paz).c3);
      _vector_sub((*paz).c4,(*pr).c4,(*paz).c4);
      paz+=1;
   }
}


static void Mj(int j,spinor *pr,spinor *pz,spinor *paz)
{
   int n,isz,isaz;

   isz=(pz-ps[0][0])/NSPIN;
   isaz=(paz-ps[0][0])/NSPIN;

   set_s2zero(VOLUME,pz);
   assign_s2s(VOLUME,pr,paz);

   for (n=0;n<ncy;n++)
      msap(nmr,isaz,isz);

   Opz(pr,paz);
}


static double wt_Msap(int nflds)
{
   int my_rank,nmax,n,i,ib;
   double wt1,wt2,wdt,wtav;
   spinor *pr,*pz,*paz;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);   
   
   pr=ps[nflds-3][0];
   pz=ps[nflds-2][0];
   paz=ps[nflds-1][0];
   Mj(0,pr,pz,paz);
   nmax=1;
   
   for (ib=0;ib<1;nmax*=2)
   { 
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();

      for (n=0;n<nmax;n++)
      {
         for (i=0;i<nflds;i+=3)
         {
            pr=ps[i][0];
            pz=ps[i+1][0];
            paz=ps[i+2][0];
            Mj(0,pr,pz,paz);
         }
      }
      
      wt2=MPI_Wtime();
      wdt=wt2-wt1;

      MPI_Reduce(&wdt,&wtav,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);      

      if (my_rank==0)
      {
         wtav/=(double)(NPROC);

         if (wtav>2.0)
            ib=1;

         wtav/=(double)((nmax*nflds)/3);
      }

      MPI_Bcast(&ib,1,MPI_INT,0,MPI_COMM_WORLD);
   }

   MPI_Bcast(&wtav,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   
   return wtav;
}


int main(int argc,char *argv[])
{
   int bs[4],nflds,n,vol;
   double wdt;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("time2.log","w",stdout);
      fin=freopen("time2.in","r",stdin);

      printf("\n");
      printf("Timing of the Schwarz preconditioner\n");
      printf("------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bs","%d %d %d %d\n",&bs[0],&bs[1],&bs[2],&bs[3]);
      read_line("nmr","%d\n",&nmr);
      read_line("ncy","%d",&ncy);
      fclose(fin);

      printf("bs = %d %d %d %d\n",bs[0],bs[1],bs[2],bs[3]);
      printf("nmr = %d\n",nmr);
      printf("ncy = %d\n\n",ncy);

      vol=bs[0]*bs[1]*bs[2]*bs[3];
      
      if (NPROC>1)
      {
         printf("There are %d MPI processes\n",NPROC);
         printf("The number of blocks per process is %d\n",VOLUME/vol);  
      }
      else
      {
         printf("There is 1 MPI process\n");
         printf("The number of blocks is %d\n",VOLUME/vol);
      }
      
      if ((VOLUME*sizeof(float))<(64*1024))
      {      
         printf("The local size of the gauge field is %d KB\n",
                (int)((72*VOLUME*sizeof(float))/(1024)));
         printf("The local size of a quark field is %d KB\n",
                (int)((24*VOLUME*sizeof(float))/(1024)));
      }
      else
      {
         printf("The local size of the gauge field is %d MB\n",
                (int)((72*VOLUME*sizeof(float))/(1024*1024)));
         printf("The local size of a quark field is %d MB\n",
                (int)((24*VOLUME*sizeof(float))/(1024*1024)));
      }

      if ((vol*sizeof(float))<(64*1024))
      {      
         printf("The size of the block gauge field is %d KB\n",
                (int)((72*vol*sizeof(float))/(1024)));
         printf("The size of a block quark field is %d KB\n",
                (int)((24*vol*sizeof(float))/(1024)));
      }
      else
      {
         printf("The size of the block gauge field is %d MB\n",
                (int)((72*vol*sizeof(float))/(1024*1024)));
         printf("The size of a block quark field is %d MB\n",
                (int)((24*vol*sizeof(float))/(1024*1024)));
      }

#if (defined SSE3)
      printf("Using inline assembly SSE3 instructions\n");
#elif (defined SSE2)
      printf("Using inline assembly SSE2 instructions\n");      
#elif (defined SSE)
      printf("Using inline assembly SSE instructions\n");
#endif

#if (defined SSE)
#if (defined P3)
      printf("Assuming SSE prefetch instructions fetch 32 bytes\n");
#elif (defined PM)
      printf("Assuming SSE prefetch instructions fetch 64 bytes\n");
#elif (defined P4)
      printf("Assuming SSE prefetch instructions fetch 128 bytes\n");
#else
      printf("SSE prefetch instructions are not used\n");
#endif
#endif
      
      printf("\n");
      fflush(flog);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmr,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncy,1,MPI_INT,0,MPI_COMM_WORLD);

   start_ranlux(0,12345);
   geometry();
#if (defined BGL)
   write_bglinfo(my_rank,"time2.bgl");
#endif
   alloc_ud();
   alloc_swd();
   alloc_bgr(GCR_BLOCKS,bs);

   set_lat_parms(0.0,0.1,0.2);
   random_ud();
   sw_term();
   assign_ud2ubgr(GCR_BLOCKS);
   n=assign_swd2swbgr(GCR_BLOCKS,ODD_PTS);
   error(n!=0,1,"main [time2.c]","Inversion of the SW term was not safe");

   nflds=(int)((4*1024*1024)/(VOLUME*sizeof(float)))+1;
   nflds+=(3-(nflds%3));
   alloc_s(nflds);

   for (n=0;n<nflds;n++)
   {
      random_s(VOLUME,ps[n][0],1.0f);
      normalize(VOLUME,1,ps[n][0]);
   }

   wdt=1.0e6*wt_Msap(nflds)/(double)(VOLUME);   
   error_chk();

   if (my_rank==0)
   {
      printf("Time per lattice point: %.2f micro sec",wdt);
      printf(" (about %d Mflops [%d bit arithmetic])\n",
             (int)((double)(24+ncy*1944+ncy*nmr*2064)/wdt),
             (int)(sizeof(spinor))/3);
      printf("Time per lattice point and MR iteration: %.2f micro sec\n\n",
             wdt/(double)(ncy*nmr));
   }

   nmr=0;
   wdt=1.0e6*wt_Msap(nflds)/(double)(VOLUME);   
   error_chk();   

   if (my_rank==0)
   {
      printf("Without block solver (i.e. setting nmr=0):\n");
      printf("Time per lattice point: %.2f micro sec",wdt);
      printf(" (about %d Mflops [%d bit arithmetic])\n",
             (int)((double)(24+ncy*1944)/wdt),(int)(sizeof(spinor))/3);
      printf("Time per lattice point and Schwarz cycle: %.2f micro sec\n\n",
             wdt/(double)(ncy));
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
