
/*******************************************************************************
*
* File block.h
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
*******************************************************************************/

#ifndef BLOCK_H
#define BLOCK_H

#ifndef SU3_H
#include "su3.h"
#endif

#ifndef START_H
#include "start.h"
#endif

#ifndef FLAGS_H
#include "flags.h"
#endif

typedef struct
{
   int ifc,ibn,vol,ns,nsd;
   int *ipp,*imb;
   su3 *u;
   su3_dble *ud;
   weyl **s;
   weyl_dble **sd;
} bndry_t;

typedef struct
{
   int *bo,*bs,vol,volh,vbb,ns,nsd,shf;
   int *ipt,*imb,(*iup)[4],(*idn)[4];
   su3 (*u)[8];
   su3_dble (*ud)[8];
   pauli *sw;
   pauli_dble *swd,*swb;
   spinor **s;
   spinor_dble **sd;
   bndry_t *bb;
} block_t;

#ifndef BLOCK_C
extern void alloc_blk(block_t *b,int *bo,int *bs,int iu,int iud,
                      int ns,int nsd);
extern void alloc_bnd(block_t *b,int ns,int nsd);
extern void free_blk(block_t *b);
extern int ipt_blk(block_t *b,int *x);
extern int dist_bndry(block_t *b,int ix);
#endif

#ifndef BLK_GRID_C
extern void alloc_bgr(blk_grid_t grid,int *bs);
extern void free_bgr(blk_grid_t grid);
extern block_t *blk_list(blk_grid_t grid,int *nb,int *isw);
#endif

#ifndef BLK_UMAP_C
extern void assign_u2ubgr(blk_grid_t grid);
extern void assign_ud2ubgr(blk_grid_t grid);
extern void assign_ud2udbgr(blk_grid_t grid);
extern void swap_active(blk_grid_t grid);
extern void assign_udb2ub(blk_grid_t grid,int n);
#endif

#ifndef BLK_SWMAP_C
extern int assign_swd2swbgr(blk_grid_t grid,ptset_t set);
extern int assign_swd2swdbgr(blk_grid_t grid,ptset_t set);
extern int assign_swdb2swb(blk_grid_t grid,int n,ptset_t set);
#endif

#ifndef BLK_SMAP_C
extern void assign_s2sb(block_t *b,ptset_t set,int k,int l);
extern void assign_sb2s(block_t *b,ptset_t set,int k,int l);
extern void assign_sd2sdb(block_t *b,ptset_t set,int k,int l);
extern void assign_sdb2sd(block_t *b,ptset_t set,int k,int l);
#endif

#endif
