
/*******************************************************************************
*
* File run3.c
*
* Copyright (C) 2005, 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Schwarz-preconditioned HMC simulation of two-flavour QCD
*
* Syntax: run3 -i <filename> [-nodfl] [-noloc] [-c <filename> [-a]]
*
* For usage instructions see the file README.run3
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "mpi.h"
#include "random.h"
#include "misc.h"
#include "start.h"
#include "sw_term.h"
#include "update.h"
#include "hmc.h"
#include "version.h"
#include "global.h"

typedef struct
{
   int nt,ns,iac;
   int stat_rop,stat_frf,stat_acf;
   int stat_gcr_frf[3],stat_gcr_acf[3];
   int stat_dfl_mds,stat_dfl_upd[2];
   double dact,avpl;
} dat_t;

static int my_rank,nodfl,noloc,scnfg,append,endian;
static int id,bc,level,seed;
static int nth,ntr,wrt_dat,ndat_buf,wrt_cnfg;
static int bs_hmc[4],N0,N1,N2,cno0,cno1,cno_blk;
static int bs_sap[4],nkv_sap,nmr_sap,ncy_sap;
static int bs_dfl[4],Ns,rinv,rnmr,rncy,rvnmx;
static int vnkv,vnmx,vdnmx;
static int nmx_rop,nmx_frf,nmx_gcr_frf;
static int nmx_acf,nmx_gcr_acf;

static double beta,kappa,csw,rkappa,tau;
static double rvres,vres,vdres;
static double res_rop,res_frf,res_gcr_frf;
static double res_acf,res_gcr_acf;

static char line[NAME_SIZE],pgrid[NAME_SIZE];
static char log_dir[NAME_SIZE],dat_dir[NAME_SIZE];
static char loc_dir[NAME_SIZE],cnfg_dir[NAME_SIZE];
static char log_file[NAME_SIZE],dat_file[NAME_SIZE];
static char log_save[NAME_SIZE],dat_save[NAME_SIZE];
static char loc_file[NAME_SIZE],cnfg_file[NAME_SIZE],end_file[NAME_SIZE];
static char base[NAME_SIZE],nbase[NAME_SIZE],cnfg[NAME_SIZE],ncnfg[NAME_SIZE];
static FILE *fin=NULL,*flog=NULL,*fdat=NULL,*fend=NULL;
#if (defined BGL)
static char bgl_file[NAME_SIZE];
#endif 

static lat_parms_t lat;
static gcr_parms_t gcr;
static hmc_parms_t hmc;
static dfl_parms_t dfl;


static int write_dat(int n,dat_t *ndat,FILE *fout)
{
   int i,iw,ic;
   stdint_t istd[15];
   double dstd[2];

   ic=0;
   
   for (i=0;i<n;i++)
   {
      istd[ 0]=(stdint_t)((*ndat).nt);
      istd[ 1]=(stdint_t)((*ndat).ns);
      istd[ 2]=(stdint_t)((*ndat).iac);   

      istd[ 3]=(stdint_t)((*ndat).stat_rop);
      istd[ 4]=(stdint_t)((*ndat).stat_frf);
      istd[ 5]=(stdint_t)((*ndat).stat_acf);      

      istd[ 6]=(stdint_t)((*ndat).stat_gcr_frf[0]);
      istd[ 7]=(stdint_t)((*ndat).stat_gcr_frf[1]);
      istd[ 8]=(stdint_t)((*ndat).stat_gcr_frf[2]);

      istd[ 9]=(stdint_t)((*ndat).stat_gcr_acf[0]);
      istd[10]=(stdint_t)((*ndat).stat_gcr_acf[1]);
      istd[11]=(stdint_t)((*ndat).stat_gcr_acf[2]);    

      istd[12]=(stdint_t)((*ndat).stat_dfl_mds);
      istd[13]=(stdint_t)((*ndat).stat_dfl_upd[0]);
      istd[14]=(stdint_t)((*ndat).stat_dfl_upd[1]);      
      
      dstd[0]=(*ndat).dact;
      dstd[1]=(*ndat).avpl;
      
      if (endian==BIG_ENDIAN)
      {
         bswap_int(15,istd);
         bswap_double(2,dstd);
      }
      
      iw =fwrite(istd,sizeof(stdint_t),15,fout);
      iw+=fwrite(dstd,sizeof(double),2,fout);

      if (iw==17)
         ic+=1;
      else
         break;
      ndat+=1;
   }
   
   return ic;
} 


static int read_dat(int n,dat_t *ndat,FILE *fin)
{
   int i,iw,ic;
   stdint_t istd[15];
   double dstd[2];

   ic=0;
   
   for (i=0;i<n;i++)
   {
      iw =fread(istd,sizeof(stdint_t),15,fin);
      iw+=fread(dstd,sizeof(double),2,fin);

      if (iw==17)
         ic+=1;
      else
         break;

      if (endian==BIG_ENDIAN)
      {
         bswap_int(15,istd);
         bswap_double(2,dstd);
      }
      
      (*ndat).nt=(int)(istd[0]);
      (*ndat).ns=(int)(istd[1]);
      (*ndat).iac=(int)(istd[2]);

      (*ndat).stat_rop=(int)(istd[3]);
      (*ndat).stat_frf=(int)(istd[4]);
      (*ndat).stat_acf=(int)(istd[5]);       

      (*ndat).stat_gcr_frf[0]=(int)(istd[ 6]);
      (*ndat).stat_gcr_frf[1]=(int)(istd[ 7]);
      (*ndat).stat_gcr_frf[2]=(int)(istd[ 8]);      

      (*ndat).stat_gcr_acf[0]=(int)(istd[ 9]);
      (*ndat).stat_gcr_acf[1]=(int)(istd[10]);
      (*ndat).stat_gcr_acf[2]=(int)(istd[11]);       

      (*ndat).stat_dfl_mds=(int)(istd[12]);
      (*ndat).stat_dfl_upd[0]=(int)(istd[13]);
      (*ndat).stat_dfl_upd[1]=(int)(istd[14]);      
      
      (*ndat).dact=dstd[0];
      (*ndat).avpl=dstd[1];

      ndat+=1;
   }
   
   return ic;
} 


static void read_infile(int argc,char *argv[])
{
   int ifile;
   
   if (my_rank==0)
   {
      flog=freopen("MPI_Errors","w",stdout);

      ifile=find_opt(argc,argv,"-i");
      nodfl=find_opt(argc,argv,"-nodfl");
      noloc=find_opt(argc,argv,"-noloc");
      scnfg=find_opt(argc,argv,"-c");
      append=find_opt(argc,argv,"-a");
      endian=endianness();
      
      error_root((ifile==0)||(ifile==(argc-1))||(scnfg==(argc-1))||
                 ((append>0)&&(scnfg==0)),1,"read_infile [run3.c]",
                 "Syntax: run3 -i <filename> "
                 "[-nodfl] [-noloc] [-c <filename> [-a]]");

      error_root(endian==UNKNOWN_ENDIAN,1,"read_infile [run3.c]",
                 "Machine has unknown endianness");
      
      if (scnfg)
      {
         strncpy(cnfg,argv[scnfg+1],NAME_SIZE-1);
         cnfg[NAME_SIZE-1]='\0';
      }
      
      fin=freopen(argv[ifile+1],"r",stdin);
      error_root(fin==NULL,1,"read_infile [run3.c]",
                 "Unable to open input file");

      read_line("id","%d\n",&id);
      read_line("log_dir","%s\n",log_dir);
      read_line("dat_dir","%s\n",dat_dir);
      read_line("loc_dir","%s\n",loc_dir);      
      read_line("cnfg_dir","%s\n",cnfg_dir);
      read_line("beta","%lf\n",&beta);
      read_line("kappa","%lf\n",&kappa);
      read_line("csw","%lf\n",&csw);
      read_line("bc","%d\n",&bc);
      read_line("level","%d\n",&level);
      read_line("seed","%d\n",&seed);

      read_line("nth","%d\n",&nth);
      read_line("ntr","%d\n",&ntr);
      read_line("wrt_dat","%d\n",&wrt_dat);
      read_line("ndat_buf","%d\n",&ndat_buf);
      read_line("wrt_cnfg","%d\n",&wrt_cnfg);

      read_line("bs_hmc","%d %d %d %d\n",
                &bs_hmc[0],&bs_hmc[1],&bs_hmc[2],&bs_hmc[3]);
      read_line("tau","%lf\n",&tau);
      read_line("N0","%d\n",&N0);
      read_line("N1","%d\n",&N1);
      read_line("N2","%d\n",&N2);
      read_line("cno0","%d\n",&cno0);
      read_line("cno1","%d\n",&cno1);
      read_line("cno_blk","%d\n",&cno_blk);
      read_line("nmx_acf","%d\n",&nmx_acf);
      read_line("res_acf","%lf\n",&res_acf);
      read_line("nmx_gcr_acf","%d\n",&nmx_gcr_acf);
      read_line("res_gcr_acf","%lf\n",&res_gcr_acf);
      read_line("nmx_rop","%d\n",&nmx_rop);
      read_line("res_rop","%lf\n",&res_rop);
      read_line("nmx_frf","%d\n",&nmx_frf);
      read_line("res_frf","%lf\n",&res_frf);
      read_line("nmx_gcr_frf","%d\n",&nmx_gcr_frf);
      read_line("res_gcr_frf","%lf\n",&res_gcr_frf);

      read_line("bs_sap","%d %d %d %d\n",
                &bs_sap[0],&bs_sap[1],&bs_sap[2],&bs_sap[3]);
      read_line("nkv_sap","%d\n",&nkv_sap);
      read_line("nmr_sap","%d\n",&nmr_sap);      
      read_line("ncy_sap","%d\n",&ncy_sap);   

      if (nodfl==0)
      {
         read_line("bs_dfl","%d %d %d %d\n",
                   &bs_dfl[0],&bs_dfl[1],&bs_dfl[2],&bs_dfl[3]);
         read_line("Ns","%d\n",&Ns);
         read_line("rkappa","%lf\n",&rkappa);
         read_line("rinv","%d\n",&rinv);
         read_line("rnmr","%d\n",&rnmr);
         read_line("rncy","%d\n",&rncy);
         read_line("rvnmx","%d\n",&rvnmx);      
         read_line("rvres","%lf\n",&rvres);
         read_line("vnkv","%d\n",&vnkv);
         read_line("vnmx","%d\n",&vnmx);
         read_line("vres","%lf\n",&vres);
         read_line("vdnmx","%d\n",&vdnmx);
         read_line("vdres","%lf",&vdres);
      }
      
      fclose(fin);
   }

   MPI_Bcast(&nodfl,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&noloc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&scnfg,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&append,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&endian,1,MPI_INT,0,MPI_COMM_WORLD);
   
   MPI_Bcast(&id,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(log_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(dat_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(loc_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(cnfg_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);

   MPI_Bcast(&beta,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&kappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&csw,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&level,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&seed,1,MPI_INT,0,MPI_COMM_WORLD);

   MPI_Bcast(&nth,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ntr,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&wrt_dat,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ndat_buf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&wrt_cnfg,1,MPI_INT,0,MPI_COMM_WORLD);

   MPI_Bcast(bs_hmc,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&tau,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&N0,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&N1,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&N2,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&cno0,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&cno1,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&cno_blk,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&nmx_acf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_acf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_gcr_acf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_gcr_acf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_rop,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_rop,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_frf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_frf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_gcr_frf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_gcr_frf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

   MPI_Bcast(bs_sap,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nkv_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmr_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncy_sap,1,MPI_INT,0,MPI_COMM_WORLD);

   if (nodfl==0)
   {
      MPI_Bcast(bs_dfl,4,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(&Ns,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rkappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&rinv,1,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(&rnmr,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rncy,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rvnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rvres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&vnkv,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&vnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&vres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&vdnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&vdres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }
}


static int check_cycle_parms(void)
{
   int nvec;
   
   nvec=start_bshift(bs_hmc);   

   error_root((nth<0)||(nth>ntr)||((nth%(ndat_buf*wrt_dat))!=0)||
              ((wrt_cnfg%(ndat_buf*wrt_dat))!=0)||(((ntr-nth)%wrt_cnfg)!=0),1,
              "check_cycle_parms [run3.c]",
              "Improper ntr, nth, wrt_dat, ndat_buf or wrt_cnfg");

   error_root(((nth%nvec)!=0)||((wrt_cnfg%nvec)!=0),1,
              "check_cycle_parms [run3.c]",
              "nth and wrt_cnfg must be multiples of the block\n"
              "shift cycle %d",nvec);

   error_root(((wrt_dat%nvec)!=0)&&((nvec%wrt_dat)!=0),1,
              "check_cycle_parms [run3.c]",
              "wrt_dat must be a multiple or an integer fraction of\n"
              "the block shift cycle %d",nvec);

   return nvec;
}


static void setup_files(void)
{
   check_dir_root(log_dir);
   check_dir_root(dat_dir);
   check_dir_root(cnfg_dir);

   if (noloc==0)
      check_dir(loc_dir);

   sprintf(base,"%dx%dx%dx%dp%db%1.2fk%.5fc%.5fid%d",
           NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,NPROC,beta,kappa,csw,id);
   sprintf(nbase,"%dx%dx%dx%db%1.2fk%.5fc%.5fid%d",
           NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);
   sprintf(pgrid,"%dx%dx%dx%d",NPROC0,NPROC1,NPROC2,NPROC3);

   if (scnfg)
   {
      MPI_Bcast(cnfg,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);

      error_root((strchr(cnfg,'n')<=strchr(cnfg,'b'))||
                 (strchr(cnfg,'b')<=strchr(cnfg,'p')),1,"setup_files [run3.c]",
                 "Incorrect configuration name %s",cnfg);
      
      strcpy(ncnfg,cnfg);

      if (strchr(cnfg,'p')!=NULL)
         strcpy(strchr(ncnfg,'p'),strchr(cnfg,'b'));
      strchr(ncnfg,'n')[0]='\0';      
   }

   error_root(((strlen(log_dir)+strlen(base)+6)>=NAME_SIZE)||
              ((strlen(dat_dir)+strlen(base)+6)>=NAME_SIZE),
              1,"setup_files [run3.c]","log_dir or dat_dir name is too long");
      
   sprintf(log_file,"%s/%s.log",log_dir,base);
   sprintf(dat_file,"%s/%s.dat",dat_dir,base);
   sprintf(end_file,"%s/%s.end",log_dir,base);
   sprintf(log_save,"%s~",log_file);
   sprintf(dat_save,"%s~",dat_file); 
#if (defined BGL)
   sprintf(bgl_file,"%s/%s.bgl",log_dir,base);
#endif
}


static void check_old_files(int *nl,int *icnfg)
{
   int log_nl,log_isv,log_icnfg,n,ir;
   char log_pgrid[NAME_SIZE];
   dat_t ndat;
   
   fend=fopen(log_file,"r");
   error_root(fend==NULL,1,"check_old_files [run3.c]",
              "Unable to open log file");

   log_pgrid[0]='\0';
   log_nl=0;
   log_isv=0;
   log_icnfg=0;
         
   for (;feof(fend)==0;)
   {
      if (fgets(line,NAME_SIZE,fend)!=NULL)
      {
         if (strstr(line,"process grid")!=NULL)
            sscanf(line,"%*s lattice, %s",log_pgrid);
         else if (strstr(line,"Trajectory no")!=NULL)
         {
            sscanf(line,"Trajectory no %d",&log_nl);
            log_isv=0;
         }
         else if (strstr(line,"Configuration no")!=NULL)
         {
            sscanf(line,"Configuration no %d",&log_icnfg);
            log_isv=1;
         }
      }
   }

   error_root(ferror(fend)!=0,1,"check_old_files [run3.c]",
              "Read error (log file)");
   fclose(fend);         

   fdat=fopen(dat_file,"rb");
   error_root(fdat==NULL,1,"check_old_files [run3.c]",
              "Unable to open data file");

   ir=0;
   
   for (n=0;feof(fdat)==0;n++)
      ir+=read_dat(1,&ndat,fdat);

   error_root((ir!=(n-1))||(ferror(fdat)!=0),1,"check_old_files [run3.c]",
              "Incorrect read count or read error (dat file)");
   fclose(fdat);   
         
   if (ir>=1)
      (*nl)=ndat.nt;
   else
      (*nl)=0;

   sscanf(strchr(cnfg,'n'),"n%d",icnfg);

   error_root(strcmp(log_pgrid,pgrid)!=0,1,"check_old_files [run3.c]",
              "Continuation run:\n"
              "Process grid does not match the previous one");

   error_root(log_icnfg!=(*icnfg),1,"check_old_files [run3.c]",
              "Continuation run:\n"
              "Initial configuration is not the last one of the previous run");

   error_root((log_isv==0)||(log_nl!=(*nl)),1,"check_old_files [run3.c]",
              "Continuation run:\n"
              "The log and/or the dat file extends beyond the last save");
         
   (*icnfg)+=1;
}


static void check_files(int *nl,int *icnfg)
{
   if (my_rank==0)
   {
      if (append)
      {
         error_root(strstr(ncnfg,nbase)==NULL,1,
                    "check_files [run3.c]","Continuation run:\n"
                    "Lattice parameters do not match the previous ones");
         
         check_old_files(nl,icnfg);
      }
      else
      {
         fend=fopen(log_file,"r");
         fdat=fopen(dat_file,"rb");

         error_root((fend!=NULL)||(fdat!=NULL),1,"check_files [run3.c]",
                    "Attempt to overwrite old *.log or *.dat file");

         error_root((scnfg)&&(strcmp(ncnfg,nbase)==0),1,"check_files [run3.c]",
                    "Run id was already used in a previous run\n"
                    "with the same lattice parameters");   

         (*nl)=0;
         (*icnfg)=1;
      }   

      if (noloc==0)
      {
         sprintf(loc_file,"/n%d_%d",(*icnfg)+(ntr-nth)/wrt_cnfg,NPROC-1);
         error_root((strlen(loc_dir)+strlen(base)+strlen(loc_file))>=NAME_SIZE,
                    1,"check_files [run3.c]","loc_dir name is too long");
      }
         
      sprintf(cnfg_file,"/n%d",(*icnfg)+(ntr-nth)/wrt_cnfg);
      error_root((strlen(cnfg_dir)+strlen(nbase)+strlen(cnfg_file))>=NAME_SIZE,
                 1,"check_files [run3.c]","cnfg_dir name is too long");
   }

   MPI_Bcast(nl,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(icnfg,1,MPI_INT,0,MPI_COMM_WORLD);
}


static void init_ud(void)
{
   if (scnfg)
   {
      if (strchr(cnfg,'p')==NULL)
      {
         sprintf(cnfg_file,"%s/%s",cnfg_dir,cnfg);
         import_cnfg(cnfg_file);
      }
      else
      {
         error_root(noloc,1,"init_ud [run3.c]","Attempt to read an "
                    "imported configuration when -noloc is set");
         
         sprintf(loc_file,"%s/%s_%d",loc_dir,cnfg,my_rank);
         read_cnfg(loc_file);
      }
   }
   else
      random_ud();
   
   error_chk();
}


static void print_info(int nvec,int nl)
{
   int ip;
   
   if (my_rank==0)
   {
      ip=ftell(flog);
      fclose(flog);

      if (ip==0)
         remove("MPI_Errors");
      
      if (append)
         flog=freopen(log_file,"a",stdout);
      else
         flog=freopen(log_file,"w",stdout);

      error_root(flog==NULL,1,"print_info [run3.c]","Unable to open log file");
      
      if (append)
         printf("Continuation run, start from configuration %s\n\n",cnfg);
      else
      {
         printf("\n");
         printf("Generation of gauge field configurations\n");
         printf("----------------------------------------\n\n");

         if (scnfg)
            printf("New run, start from configuration %s\n\n",cnfg);
         else
            printf("New run, start from random configuration\n\n");
      }

      printf("Program version %s\n",DD_HMC_RELEASE);

      if (endian==LITTLE_ENDIAN)
         printf("The machine is little endian\n");
      else
         printf("The machine is big endian\n");
      if (noloc)
         printf("The local disks are not used\n");
      if (nodfl)
         printf("Deflation is switched off\n");
         
      printf("\n");      
      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      printf("O(a) improved theory\n");
      printf("beta = %.6f\n",lat.beta);
      printf("kappa = %.6f\n",lat.kappa);
      printf("csw = %.6f\n",lat.csw);
      printf("bc = %+d\n",hmc.bc);

      if ((scnfg)&&(strchr(cnfg,'p')!=NULL))
      {
         printf("State of ranlxs and ranlxd read from\n");
         printf("initial field-configuration file\n\n");
      }
      else
         printf("level = %d, seed = %d, effective seed = %d\n\n",
                level,seed,seed^nl);

      printf("nth = %d, ntr = %d\n",nth,ntr);
      printf("wrt_dat = %d, ndat_buf = %d, wrt_cnfg = %d\n\n",
             wrt_dat,ndat_buf,wrt_cnfg);

      printf("bs_hmc = %d %d %d %d, ",
             bs_hmc[0],bs_hmc[1],bs_hmc[2],bs_hmc[3]);
      printf("block shift cycle = %d\n",nvec);
      printf("tau = %.6f\n",hmc.tau);
      printf("N0,N1,N2 = %d,%d,%d\n",hmc.N0,hmc.N1,hmc.N2);
      printf("cno0 = %d, cno1 = %d, cno_blk = %d\n",
             hmc.cno0,hmc.cno1,hmc.cno_blk);
      printf("nmx_rop = %d, res_rop = %.1e\n",hmc.nmx_rop,hmc.res_rop);
      printf("nmx_acf = %d, res_acf = %.1e\n",hmc.nmx_acf,hmc.res_acf);
      printf("nmx_gcr_acf = %d, res_gcr_acf = %.1e\n",
             hmc.nmx_gcr_acf,hmc.res_gcr_acf);
      printf("nmx_frf = %d, res_frf = %.1e\n",hmc.nmx_frf,hmc.res_frf);
      printf("nmx_gcr_frf = %d, res_gcr_frf = %.1e\n\n",
             hmc.nmx_gcr_frf,hmc.res_gcr_frf);
      
      printf("bs_sap = %d %d %d %d\n",
             bs_sap[0],bs_sap[1],bs_sap[2],bs_sap[3]);
      printf("nkv_sap,nmr_sap,ncy_sap = %d,%d,%d\n\n",
             gcr.nkv,gcr.nmr,gcr.ncy);

      if (nodfl==0)
      {
         printf("bs_dfl = %d %d %d %d\n",
                bs_dfl[0],bs_dfl[1],bs_dfl[2],bs_dfl[3]);
         printf("Ns = %d\n",dfl.Ns);
         printf("rkappa = %.6f\n",dfl.rkappa);
         printf("rinv = %d, rnmr = %d, rncy = %d\n",dfl.rinv,dfl.rnmr,dfl.rncy);
         printf("rvnmx = %d, rvres = %.1e\n",dfl.rvnmx,dfl.rvres);
         printf("vnkv = %d\n",dfl.vnkv);         
         printf("vnmx = %d, vres = %.1e\n",dfl.vnmx,dfl.vres);
         printf("vdnmx = %d, vdres = %.1e\n\n",dfl.vdnmx,dfl.vdres);
      }

      fflush(flog);
   }
}


static void print_log(dat_t *ndat)
{
   if (my_rank==0)
   {
      printf("Trajectory no %d\n",(*ndat).nt);

      printf("start_md:\n");
      printf("nshift = %d, ",(*ndat).ns);
      if (nodfl)
         printf("stat_rop = %d\n",(*ndat).stat_rop);
      else
      {
         printf("stat_rop = %d, ",(*ndat).stat_rop);
         printf("stat_dfl_mds = %d\n",(*ndat).stat_dfl_mds);
      }
      
      printf("run_md:\n");
      printf("stat_frf = %d, ",(*ndat).stat_frf);
      if (nodfl)
         printf("stat_gcr_frf = %d\n",(*ndat).stat_gcr_frf[0]);
      else
      {
         printf("stat_gcr_frf = %d (little = %d, %d), ",
                (*ndat).stat_gcr_frf[0],(*ndat).stat_gcr_frf[1],
                (*ndat).stat_gcr_frf[2]);
         printf("stat_dfl_upd = %d (little = %d)\n",(*ndat).stat_dfl_upd[0],
                (*ndat).stat_dfl_upd[1]);         
      }
      printf("stat_acf = %d, ",(*ndat).stat_acf);
      if (nodfl)
         printf("stat_gcr_acf = %d\n",(*ndat).stat_gcr_acf[0]);
      else
         printf("stat_gcr_acf = %d (little = %d, %d)\n",
                (*ndat).stat_gcr_acf[0],(*ndat).stat_gcr_acf[1],
                (*ndat).stat_gcr_acf[2]);

      printf("accept_step:\n");
      printf("dH = %+.1e, ",(*ndat).dact);
      printf("iac = %d\n",(*ndat).iac);
      printf("Average plaquette = %.6f\n\n",(*ndat).avpl);
   }
}


static void save_dat(int n,int siac,double wtcyc,double wtall,dat_t *adat)
{
   int iw;
   
   if (my_rank==0)
   {
      fdat=fopen(dat_file,"ab");
      error_root(fdat==NULL,1,"save_dat [run3.c]",
                 "Unable to open data file");

      iw=write_dat(ndat_buf,adat,fdat);      

      error_root((iw!=ndat_buf)||(ferror(fdat)!=0),1,"save_dat [run3.c]",
                 "Error while saving data to the *.dat file");
      fclose(fdat);

      printf("%d data items saved\n",ndat_buf);
      printf("Acceptance rate = %1.2f\n",
             (double)(siac)/(double)(n+1));

      printf("Time per trajectory: last cycle = %.3e sec, "
             "all cycles = %.3e sec\n",wtcyc/(double)(wrt_dat*ndat_buf),
             wtall/(double)(n+1));
      
      if (((n+1)<nth)||(((ntr-n-1)%wrt_cnfg)!=0))
         printf("\n");
   }
}


static void save_cnfg(int icnfg,int warn)
{
   if (noloc==0)
   {
      sprintf(loc_file,"%s/%sn%d_%d",loc_dir,base,icnfg,my_rank);
      write_cnfg(loc_file);
   }
         
   sprintf(cnfg_file,"%s/%sn%d",cnfg_dir,nbase,icnfg);
   export_cnfg(cnfg_file);

   if (my_rank==0)
   {
      if (noloc==0)
         printf("Configuration no %d saved on the local disks "
                "and exported\n\n",icnfg);
      else
         printf("Configuration no %d exported\n\n",icnfg);

      if (warn)
         printf("WARNING: the inner solver for the little system failed in\n"
                "%d cases. For best performance, suggest to increase the\n"
                "parameter vnmx in the input file\n\n",warn);
   }
}


static void check_endflag(int *iend)
{
   if (my_rank==0)
   {
      fend=fopen(end_file,"r");

      if (fend!=NULL)
      {
         fclose(fend);
         remove(end_file);
         (*iend)=1;
         printf("End flag set, run stopped\n\n");
      }
   }

   MPI_Bcast(iend,1,MPI_INT,0,MPI_COMM_WORLD);
}


int main(int argc,char *argv[])
{
   int nvec,nl,icnfg;
   int idat,iend,warn,siac;
   int n,ns,ifail,iac;
   int stat_rop,stat_frf,stat_gcr_frf[3];
   int stat_acf,stat_gcr_acf[3],stat_dfl_mds,stat_dfl_upd[2];
   double dact,avpl;
   double wt1,wt2,wtcyc,wtall;
   hmc_act_t act0,act1;
   dat_t *adat,*ndat;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   read_infile(argc,argv);
   nvec=check_cycle_parms();
   setup_files();
   check_files(&nl,&icnfg);

   start_ranlux(level,seed^nl);
   geometry();
#if (defined BGL)
   write_bglinfo(my_rank,bgl_file);
#endif 
   alloc_ud();
   alloc_swd();

   lat=set_lat_parms(beta,kappa,csw);
   gcr=set_gcr_parms(bs_sap,nmr_sap,ncy_sap,nkv_sap);
   hmc=set_hmc_parms(bs_hmc,bc,N0,N1,N2,nodfl==0,cno0,cno1,cno_blk,
                     nmx_acf,nmx_gcr_acf,nmx_rop,nmx_frf,nmx_gcr_frf,
                     tau,
                     res_acf,res_gcr_acf,res_rop,res_frf,res_gcr_frf);

   if (nodfl==0)
   {
      if ((cno0+cno1)==0)
         alloc_sd(7);
      else
         alloc_sd(8);
      
      if (Ns<(2*nkv_sap))
         alloc_s(2*nkv_sap+1);
      else
         alloc_s(Ns+2);

      alloc_u();
      alloc_sw();
      dfl=set_dfl_parms(bs_dfl,Ns,rinv,rnmr,rncy,rvnmx,vnkv,vnmx,vdnmx,
                        rkappa,rvres,vres,vdres,0,0);
   }
   else
   {
      if ((cno0+cno1)==0)
         alloc_sd(5);
      else
         alloc_sd(6);
      
      alloc_s(2*nkv_sap+1);
   }

   alloc_hmc_action(&act0);
   alloc_hmc_action(&act1);
   adat=amalloc(ndat_buf*sizeof(dat_t),3);
   error_root(adat==NULL,1,"main [run3.c]","Unable to allocate data array");
   
   init_ud();
   print_info(nvec,nl);

   idat=0;
   iend=0;
   warn=0;
   siac=0;
   wtcyc=0.0;
   wtall=0.0;

   for (n=0;(iend==0)&&(n<ntr);n++)
   {
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();      
      ns=mvfld();
      start_md(act0,&stat_rop,&stat_dfl_mds,&ifail);

      error_root((stat_rop<0)||(stat_dfl_mds<0)||(ifail!=0),1,"main [run3.c]",
                 "start_md() failed: stat_rop=%d, stat_dfl_mds=%d, ifail=%d",
                 stat_rop,stat_dfl_mds,ifail);

      run_md(act1,&stat_acf,stat_gcr_acf,&stat_frf,stat_gcr_frf,
             stat_dfl_upd,&ifail);

      error_root((stat_acf<0)||(stat_gcr_acf[0]<0)||(stat_gcr_acf[2]<0)||
                 (stat_frf<0)||(stat_gcr_frf[0]<0)||(stat_gcr_frf[2]<0)||
                 (stat_dfl_upd[1]<0)||(ifail!=0),1,"main [run3.c]",
                 "run_md() failed: stat_acf=%d, stat_gcr_acf=%d,%d,%d\n"
                 "                 stat_frf=%d, stat_gcr_frf=%d,%d,%d\n"
                 "                 stat_dfl_upd=%d,%d, ifail=%d",
                 stat_acf,stat_gcr_acf[0],stat_gcr_acf[1],stat_gcr_acf[2],
                 stat_frf,stat_gcr_frf[0],stat_gcr_frf[1],stat_gcr_frf[2],
                 stat_dfl_upd[0],stat_dfl_upd[1],ifail);
                 
      iac=accept_step(act0,act1,&dact);
      MPI_Barrier(MPI_COMM_WORLD);
      wt2=MPI_Wtime();        

      warn+=(stat_gcr_acf[1]<0)+(stat_gcr_frf[1]<0);
      siac+=iac;
      wtcyc+=(wt2-wt1);
      
      if (((ntr-n-1)%wrt_dat)==0)
      {
         avpl=plaq_sum_dble()/(double)(6*VOLUME*NPROC);

         if (my_rank==0)
         {
            ndat=adat+idat;
            (*ndat).nt=nl+n+1;
            (*ndat).ns=ns;
            (*ndat).iac=iac;
            (*ndat).stat_rop=stat_rop;
            (*ndat).stat_dfl_mds=stat_dfl_mds;
            (*ndat).stat_frf=stat_frf;
            (*ndat).stat_gcr_frf[0]=stat_gcr_frf[0];
            (*ndat).stat_gcr_frf[1]=stat_gcr_frf[1];
            (*ndat).stat_gcr_frf[2]=stat_gcr_frf[2];            
            (*ndat).stat_acf=stat_acf;
            (*ndat).stat_gcr_acf[0]=stat_gcr_acf[0];
            (*ndat).stat_gcr_acf[1]=stat_gcr_acf[1];
            (*ndat).stat_gcr_acf[2]=stat_gcr_acf[2];            
            (*ndat).stat_dfl_upd[0]=stat_dfl_upd[0];
            (*ndat).stat_dfl_upd[1]=stat_dfl_upd[1];

            (*ndat).dact=dact;
            (*ndat).avpl=avpl;

            print_log(ndat);
            idat+=1;

            if (idat==ndat_buf)
            {
               wtall+=wtcyc;
               save_dat(n,siac,wtcyc,wtall,adat);
               idat=0;
               wtcyc=0.0;
            }

            fflush(flog);
         }
      }

      if (((n+1)>=nth)&&(((ntr-n-1)%wrt_cnfg)==0))
      {
         save_cnfg(icnfg,warn);
         check_endflag(&iend);
         error_chk();
         icnfg+=1;

         if (my_rank==0)
         {
            fflush(flog);
            copy_file(log_file,log_save);
            copy_file(dat_file,dat_save);
         }
      }
   }

   if (my_rank==0)
      fclose(flog);

   MPI_Finalize();
   exit(0);
}
